/* 
 *	Copyright (C) Chia-chen Kuo - April 2001
 *
 *  This file is part of DVD2AVIMFC, a free MPEG-2 decoder
 *	
 *  DVD2AVIMFC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVD2AVIMFC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "global.h"
#include "getbit.h"
#include "AC3Dec\ac3.h"

#define LOCATE												\
while (Rdptr >= (Rdbfr + BUFFER_SIZE))						\
{															\
	Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);	\
	if (Read < BUFFER_SIZE) Next_File();					\
	Rdptr -= BUFFER_SIZE;									\
}

#define DECODE_AC3																\
{																				\
	size = 0;																	\
	while (Packet_Length > 0)													\
	{																			\
		if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
		{																		\
			size = ac3_decode_data(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, size);		\
			Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
			Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
			if (Read < BUFFER_SIZE) Next_File();								\
			Rdptr = Rdbfr;														\
		}																		\
		else																	\
		{																		\
			size = ac3_decode_data(Rdptr, Packet_Length, size);					\
			Rdptr += Packet_Length;												\
			Packet_Length = 0;													\
		}																		\
	}																			\
}

#define DEMUX_AC3															\
while (Packet_Length > 0)													\
{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, ac3[This_Track].file);\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
		Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
		if (Read < BUFFER_SIZE) Next_File();								\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, Packet_Length, 1, ac3[This_Track].file);			\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}

#define DEMUX_PCM														\
{																		\
	unsigned char tmp;													\
	size = 0;															\
	while (Packet_Length > 0)											\
	{																	\
		if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)					\
		{																\
			memcpy(PCM_Buffer+size, Rdptr, BUFFER_SIZE+Rdbfr-Rdptr);	\
			size += BUFFER_SIZE+Rdbfr-Rdptr;							\
			Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;					\
			Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);		\
			if (Read < BUFFER_SIZE) Next_File();						\
			Rdptr = Rdbfr;												\
		}																\
		else															\
		{																\
			memcpy(PCM_Buffer+size, Rdptr, Packet_Length);				\
			Rdptr += Packet_Length;										\
			size += Packet_Length;										\
			Packet_Length = 0;											\
		}																\
	}																	\
	for (i=0; i<size; i+=2)												\
	{																	\
		tmp = PCM_Buffer[i];											\
		PCM_Buffer[i] = PCM_Buffer[i+1];								\
		PCM_Buffer[i+1] = tmp;											\
	}																	\
}

#define DEMUX_MPA															\
while (Packet_Length > 0)													\
{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, mpa[MPA_Track].file);	\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
		Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
		if (Read < BUFFER_SIZE) Next_File();								\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, Packet_Length, 1, mpa[MPA_Track].file);			\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}

#define DEMUX_DTS															\
while (Packet_Length > 0)													\
{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, dts[This_Track].file);\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
		Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
		if (Read < BUFFER_SIZE) Next_File();								\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, Packet_Length, 1, dts[This_Track].file);			\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}

static char *FTType[5] = {
	"48KHz", "44.1KHz", "44.1KHz", "44.1KHz", "44.1KHz"
};

static char *AC3ModeDash[8] = {
	"1+1", "1_0", "2_0", "3_0", "2_1", "3_1", "2_2", "3_2"
};

static char *AC3Mode[8] = {
	"1+1", "1/0", "2/0", "3/0", "2/1", "3/1", "2/2", "3/2"
};

static int AC3Rate[32] = {
	32, 40, 48, 56, 64, 80, 96, 112, 128, 160,
	192, 224, 256, 320, 384, 448, 512, 576, 640,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

// JEANL: Making this guys non-static so we can read them from elsewhere...
int VOB_ID, CELL_ID;
int NextVOB_ID, NextCELL_ID;
char Num_Buttons, NextNum_Buttons;
int VideoPTS, AudioPTS;
static unsigned char PCM_Buffer[BUFFER_SIZE];
static short *ptrPCM_Buffer = (short*)PCM_Buffer;

int DrawButtons = 0;		// If 1, the button info is read, and the button rectangles drawn.
int ButtonsChanged = 1;		// Indicates button info did change since last data.
int StartVobNumber = 0;		// Number of first VOB in our array of files.
unsigned short NextStartY[36]; 
unsigned short NextEndY[36];	
unsigned short NextStartX[36];
unsigned short NextEndX[36];

unsigned short StartY[36]; 
unsigned short EndY[36];	
unsigned short StartX[36];
unsigned short EndX[36];


void Initialize_Buffer()
{
	Rdptr = Rdbfr + BUFFER_SIZE;
	Rdmax = Rdptr;

	if (SystemStream_Flag)
	{
		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr = *Rdptr++ << 24;

		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr += *Rdptr++ << 16;

		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr += *Rdptr++ << 8;

		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr += *Rdptr++;

		Fill_Next();
	}
	else
	{
		Fill_Buffer();

		CurrentBfr = (*Rdptr << 24) + (*(Rdptr+1) << 16) + (*(Rdptr+2) << 8) + *(Rdptr+3);
		Rdptr += 4;

		Fill_Next();
	}

	BitsLeft = 32;
}

void Next_Packet()
{
	static int i, Packet_Length, Packet_Header_Length, size;
	static unsigned int code, AUDIO_ID, VOBCELL_Count, This_Track, MPA_Track;

	for (;;)
	{
		code = Get_Short();
		code = (code<<16) + Get_Short();

		while ((code & 0xffffff00) != 0x00000100)
			code = (code<<8) + Get_Byte();

		switch (code)
		{
			case PACK_START_CODE:

				Rdptr += 8;
				VOBCELL_Count = 0;
				break;

			case PRIVATE_STREAM_2:
				Packet_Length = Get_Short();

				if (++VOBCELL_Count==2)
				{
					// JEANL: I modified this code because the VOB/CELL ID was inaccurate in some cases:
					// Sometimes, the pic was decoded OK, but then filling the next buffer caused the Nav
					// Pack to be parsed, and the VOB/CELL ID to be saved, even though they don't correspond
					// to the previous image. Here, I'm saving them as "Next" values. decode_image() will
					// update the fields and the true value based on these.
					Rdptr += 25;
					NextVOB_ID = Get_Short();
					// Was Get_Short(); But that was incorrect! The cellID is a byte, and the value next to it
					// is "reserved", 0 most of the time, but sometimes not!
					Rdptr ++;
					NextCELL_ID = Get_Byte();
					Rdptr += Packet_Length - 29;
				}
				else
				{
					// JEANL: Adding button count detection! 
					// That's in the first PRIVATE_STREAM_2 packet.
					// The number of buttons is at offset 114 after the Length.
					char LastNumButtons = NextNum_Buttons;
					Rdptr += 114;
					NextNum_Buttons = Get_Byte();

#if 1
					// This code reads and parses the button info... They will be displayed in store.c
					// Note that I fill out the Next* arrays. The Current arrays are copied in Decode_Picture()
					// (in getpic.c).
					Rdptr += 0x1c;
					// ButtonsChanged will be reset by the function that draws the buttons.
					ButtonsChanged |= LastNumButtons != NextNum_Buttons;
//OutputDebugString("GetBit\n"); //Use this to see when this function is called.
					for(i=0;i<NextNum_Buttons;i++)
					{
						unsigned char Byte1 = Get_Byte();
						unsigned char Byte2 = Get_Byte();
						unsigned char Byte3 = Get_Byte();
						unsigned char Byte4 = Get_Byte();
						unsigned char Byte5 = Get_Byte();
						unsigned char Byte6 = Get_Byte();
						unsigned short LastVal;

						LastVal = NextStartX[i];
						NextStartX[i] = ((unsigned short)(Byte1 & 0x3f)) << 4; 
						NextStartX[i] |= ((unsigned short)(Byte2 & 0xf0)) >> 4;
						ButtonsChanged |= LastVal != NextStartX[i];
						LastVal = NextEndX[i];
						NextEndX[i] = ((unsigned short)(Byte2 & 0x03)) << 8; 
						NextEndX[i] |= (unsigned short)(Byte3);
						ButtonsChanged |= LastVal != NextEndX[i];

						LastVal = NextStartY[i];
						NextStartY[i] = ((unsigned short)(Byte4 & 0x3f)) << 4; 
						NextStartY[i] |= ((unsigned short)(Byte5 & 0xf0)) >> 4;
						ButtonsChanged |= LastVal != NextStartY[i];
						LastVal = NextEndY[i];
						NextEndY[i] = ((unsigned short)(Byte5 & 0x03)) << 8; 
						NextEndY[i] |= (unsigned short)(Byte6);
						ButtonsChanged |= LastVal != NextEndY[i];

						Rdptr+= 12;
					}
					// Reset point back to where we were right after reading NextNum_Buttons 
					Rdptr -= (0x1c + NextNum_Buttons*18);
#endif

					// Now reset pointer where we were, and resume the normal original code.
					Rdptr -= 115;
					Rdptr += Packet_Length;
				}
				break;

			case PRIVATE_STREAM_1:
				Packet_Length = Get_Short();

				Rdptr ++;	// +1
				code = Get_Byte();	// +1
				Packet_Header_Length = Get_Byte();	// +1

				if (code>=0x80)
				{
					code = Get_Byte();
					AudioPTS = (code & 0x0e) << 29;

					AudioPTS |= (Get_Short() & 0xfffe) << 14;
					AudioPTS |= (Get_Short()>>1) & 0x7fff;

					Rdptr += Packet_Header_Length-5;
				}
				else
				{
					// JEANL: If there's no PTS, an old value of AudioPTS will be displayed, that's
					// really bad when there's timing discontinuity. I'm setting this to 0.
					AudioPTS = 0;
					Rdptr += Packet_Header_Length;
				}

				AUDIO_ID = Get_Byte();	// +1
				Packet_Length -= Packet_Header_Length+4;

				if (AUDIO_ID>=SUB_AC3 && AUDIO_ID<SUB_AC3+CHANNEL)
				{
					Rdptr += 3; Packet_Length -= 3; This_Track = AUDIO_ID-SUB_AC3;

					LOCATE

					if (!ac3[This_Track].rip && Start_Flag && !Channel[This_Track])
					{
						Channel[This_Track] = FORMAT_AC3;

						code = Get_Byte();
						code = (code & 0xff)<<8 | Get_Byte();
						i = 0;

						while (code!=0xb77)
						{
							code = (code & 0xff)<<8 | Get_Byte();
							i++;
						}

						Get_Short();
						ac3[This_Track].rate = (Get_Byte()>>1) & 0x1f;
						Get_Byte();
						ac3[This_Track].mode = (Get_Byte()>>5) & 0x07;

						Rdptr -= 7; Packet_Length -= i;

						if ((AVI_Flag || D2V_Flag || Decision_Flag))
						{
							if (Decision_Flag && This_Track==Track_Flag)
							{
								InitialAC3();

								DECODE_AC3

								ac3[This_Track].rip = true;
							}
							else if (Method_Flag==AUDIO_DECODE && This_Track==Track_Flag)
							{
								sprintf(szBuffer, "%s AC3 T%02d %sch %dKbps %s.wav", szOutput, This_Track+1, 
									AC3ModeDash[ac3[This_Track].mode], AC3Rate[ac3[This_Track].rate], FTType[SRC_Flag]);

								strcpy(pcm.filename, szBuffer);
								pcm.file = fopen(szBuffer, "wb");

								StartWAV(pcm.file);
								pcm.delay = ((AudioPTS-VideoPTS)/90) * 192;

								if (SRC_Flag)
								{
									DownWAV(pcm.file);
									InitialSRC();
								}

								if (pcm.delay > 0)
								{
									if (SRC_Flag)
										pcm.delay = ((int)(0.91875*pcm.delay)>>2)<<2;

									for (i=0; (i<pcm.delay) && (Dry == false); i++)
										fputc(0, pcm.file);

									pcm.size += pcm.delay;
									pcm.delay = 0;
								}

								InitialAC3();

								DECODE_AC3

								if (-pcm.delay > size)
									pcm.delay += size;
								else
								{
									if (SRC_Flag)
										Wavefs44(pcm.file, size+pcm.delay, AC3Dec_Buffer-pcm.delay);
									else
										fwrite(AC3Dec_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);

									pcm.size += size+pcm.delay;
 									pcm.delay = 0;
								}

								ac3[This_Track].rip = true;
							}
							else if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && This_Track==Track_Flag)
							{
								sprintf(szBuffer, "%s AC3 T%02d %sch %dKbps DELAY %dms.ac3", szOutput, This_Track+1, 
									AC3ModeDash[ac3[This_Track].mode], AC3Rate[ac3[This_Track].rate], (AudioPTS-VideoPTS)/90);

								ac3[This_Track].file = fopen(szBuffer, "wb");

								DEMUX_AC3

								ac3[This_Track].rip = true;
							}
						}
					}
					else if (ac3[This_Track].rip)
					{
						if (Decision_Flag)
							DECODE_AC3
						else if (Method_Flag==AUDIO_DECODE)
						{
							DECODE_AC3

							if (-pcm.delay > size)
								pcm.delay += size;
							else
							{
								if (SRC_Flag)
									Wavefs44(pcm.file, size+pcm.delay, AC3Dec_Buffer-pcm.delay);
								else if (Dry == false)
									fwrite(AC3Dec_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);

								if (Dry == false)
									pcm.size += size+pcm.delay;
 								pcm.delay = 0;
							}
						}
						else
							DEMUX_AC3
					}
				}
				else if (AUDIO_ID>=SUB_PCM && AUDIO_ID<SUB_PCM+CHANNEL)
				{
					Rdptr += 6; Packet_Length -= 6; This_Track = AUDIO_ID-SUB_PCM;

					LOCATE

					if (!pcm.rip && Start_Flag && !Channel[This_Track])
					{
						Channel[This_Track] = FORMAT_LPCM;

						if (AVI_Flag || D2V_Flag)
						{
							if (Method_Flag==AUDIO_DECODE && This_Track==Track_Flag)
							{
								pcm.delay = ((AudioPTS-VideoPTS)/90) * 192;

								sprintf(szBuffer, "%s LPCM T%02d %s.wav", szOutput, This_Track+1, FTType[SRC_Flag]);
								strcpy(pcm.filename, szBuffer);

								pcm.file = fopen(szBuffer, "wb");
								StartWAV(pcm.file);

								if (SRC_Flag)
								{
									DownWAV(pcm.file);
									InitialSRC();
								}

								if (pcm.delay > 0)
								{
									if (SRC_Flag)
										pcm.delay = ((int)(0.91875*pcm.delay)>>2)<<2;

									for (i=0; (i<pcm.delay) && (Dry == false); i++)
										fputc(0, pcm.file);

									pcm.size += pcm.delay;
									pcm.delay = 0;
								}

								if (-pcm.delay > Packet_Length)
									pcm.delay += Packet_Length;
								else
								{
									DEMUX_PCM

									if (SRC_Flag)
										Wavefs44(pcm.file, size+pcm.delay, PCM_Buffer-pcm.delay);
									else
										fwrite(PCM_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);
	
									pcm.size += size+pcm.delay;
									pcm.delay = 0;
								}

								pcm.rip = true;
							}
						}
					}
					else if (pcm.rip)
					{
						if (-pcm.delay > Packet_Length)
							pcm.delay += Packet_Length;
						else
						{
							DEMUX_PCM

							if (SRC_Flag)
								Wavefs44(pcm.file, size+pcm.delay, PCM_Buffer-pcm.delay);
							else
							{
								if (Dry == false)
									fwrite(PCM_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);
								
								if (Norm_Flag)
									for (i=0; i<(size>>1); i++)
										if (Sound_Max < abs(ptrPCM_Buffer[i]))
											Sound_Max = abs(ptrPCM_Buffer[i]);
							}

							if (Dry == false)
								pcm.size += size+pcm.delay;
							pcm.delay = 0;
						}
					}
				}
				else if (AUDIO_ID>=SUB_DTS && AUDIO_ID<SUB_DTS+CHANNEL)
				{
					Rdptr += 3; Packet_Length -= 3; This_Track = AUDIO_ID-SUB_DTS;

					LOCATE

					if (!dts[This_Track].rip && Start_Flag && !Channel[This_Track])
					{
						Channel[This_Track] = FORMAT_DTS;

						if (AVI_Flag || D2V_Flag)
						{
							if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && This_Track==Track_Flag)
							{
								sprintf(szBuffer, "%s DTS T%02d DELAY %dms.dts", szOutput, This_Track+1, (AudioPTS-VideoPTS)/90);

								dts[This_Track].file = fopen(szBuffer, "wb");

								DEMUX_DTS

								dts[This_Track].rip = true;
							}
						}
					}
					else if (dts[This_Track].rip)
						DEMUX_DTS
				}
				Rdptr += Packet_Length;
				break;

			case AUDIO_ELEMENTARY_STREAM_7:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_6:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_5:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_4:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_3:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_2:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_1:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_0:
				Packet_Length = Get_Short()-1;
				code = Get_Byte();

				if ((code & 0xc0)==0x80)
				{
					code = Get_Byte();	// +1
					Packet_Header_Length = Get_Byte();	// +1

					if (code>=0x80)
					{
						code = Get_Byte();

						AudioPTS = (code & 0x0e) << 29;
						AudioPTS |= (Get_Short() & 0xfffe) << 14;
						AudioPTS |= (Get_Short()>>1) & 0x7fff;

						Rdptr += Packet_Header_Length-5;
					}
					else
						Rdptr += Packet_Header_Length;

					Packet_Length -= Packet_Header_Length+2;

					LOCATE

					if (!mpa[MPA_Track].rip && Start_Flag && !Channel[MPA_Track])
					{
						Channel[MPA_Track] = FORMAT_MPA;

						code = Get_Byte();
						code = (code & 0xff)<<8 | Get_Byte();
						i = 0;

						while (code<0xfff0)
						{
							code = (code & 0xff)<<8 | Get_Byte();
							i++;
						}

						Rdptr -= 2; Packet_Length -= i;

						if ((AVI_Flag || D2V_Flag))
						{
							if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && MPA_Track==Track_Flag)
							{
								sprintf(szBuffer, "%s MPA T%02d DELAY %dms.mpa", szOutput, MPA_Track+1, (AudioPTS-VideoPTS)/90);
								mpa[MPA_Track].file = fopen(szBuffer, "wb");

								DEMUX_MPA

								mpa[MPA_Track].rip = true;
							}
						}
					}
					else if (mpa[MPA_Track].rip)
						DEMUX_MPA
				}

				Rdptr += Packet_Length;
				MPA_Track = 0;
				break;

			case VIDEO_ELEMENTARY_STREAM:
				Packet_Length = Get_Short();
				Rdmax = Rdptr + Packet_Length;

				code = Get_Byte();

				if ((code & 0xc0)==0x80)
				{
					code = Get_Byte();
					Packet_Header_Length = Get_Byte();

					if (code>=0x80 && !Start_Flag)
					{
						code = Get_Byte();
						VideoPTS = (code & 0x0e) << 29;
						VideoPTS |= (Get_Short() & 0xfffe) << 14;
						VideoPTS |= (Get_Short()>>1) & 0x7fff;
						// Adjust VideoPTS as required if any B frames precede the first I frame
						// in display order.
						if (Frame_Rate)
							VideoPTS -= (int) ((LeadingBFrames * 90000) / Frame_Rate);

						Rdptr += Packet_Header_Length-5;
					}
					else
						Rdptr += Packet_Header_Length;

					Bitrate_Monitor += Rdmax-Rdptr;
					return;
				}
				else
				{
					Rdptr += Packet_Length-1;
				}

				break;

			default:
				if (code>=SYSTEM_START_CODE)
				{
					Packet_Length = Get_Short();
					Rdptr += Packet_Length;
				}
				break;
		}
	}
}

unsigned int Get_Bits_All(unsigned int N)
{
	N -= BitsLeft;
	Val = (CurrentBfr << (32 - BitsLeft)) >> (32 - BitsLeft);

	if (N)
		Val = (Val << N) + (NextBfr >> (32 - N));

	CurrentBfr = NextBfr;
	BitsLeft = 32 - N;
	Fill_Next();

	return Val;
}

void Flush_Buffer_All(unsigned int N)
{
	CurrentBfr = NextBfr;
	BitsLeft = BitsLeft + 32 - N;
	Fill_Next();
}

void Fill_Buffer()
{
	Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);

	if (Read < BUFFER_SIZE)
		Next_File();

	if (KeyOp_Flag && (Rdbfr[20] & 0x10))
	{
		BufferOp(Rdbfr, lfsr0, lfsr1);
		Rdbfr[20] &= ~0x10;
	}

	Rdptr = Rdbfr;

	if (SystemStream_Flag)
		Rdmax -= BUFFER_SIZE;
	else
		Bitrate_Monitor += Read;
}

void Next_File()
{
	int i;
	
	if (File_Flag < File_Limit-1) File_Flag++;
	else
	{
		// We can't do a thread kill here because we'll lose
		// the last two frames. Set up to kill at the next Get_Hdr()
		// call. We allow the last file to wrap around but we won't
		// use any of it after wrapping. I know, it's kludgy, but it works.
		if(Dry)
		{
			// JEANL: I'm not sure I understand the meaning of the Dry variable, but it seems
			// that if we try to open an invalid VOB file (encrypted for example) or a file of
			// some other type, we end up here. We don't if the vob files are valid.
			Fault_Flag = CRITICAL_ERROR_LEVEL;
			ThreadKill();
		}
		Dry = true;
	}

	process.run = 0;
	for (i=0; i<File_Flag; i++)
		process.run += process.length[i];

	_lseeki64(Infile[File_Flag], 0, SEEK_SET);
	_read(Infile[File_Flag], Rdbfr + Read, BUFFER_SIZE - Read);
}

void UpdateInfo()
{
	static int Old_NTSC_Purity;
	int pts;

	sprintf(szBuffer, "%s", AspectRatio[aspect_ratio_information]);
	SetDlgItemText(hDlg, IDC_ASPECT_RATIO, szBuffer);

	sprintf(szBuffer, "%.3f fps", Frame_Rate);
	SetDlgItemText(hDlg, IDC_FRAME_RATE, szBuffer);

	switch (Channel[Track_Flag])
	{
		case FORMAT_AC3:
			sprintf(szBuffer, "DD %s %d", AC3Mode[ac3[Track_Flag].mode], AC3Rate[ac3[Track_Flag].rate]);
			SetDlgItemText(hDlg, IDC_AUDIO_TYPE, szBuffer);
			break;

		case FORMAT_MPA:
			SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "MPEG Audio");
			break;

		case FORMAT_LPCM:
			SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "Linear PCM");
			break;

		case FORMAT_DTS:
			SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "DTS");
			break;

		default:
			SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "");
			break;
	}

	if (SystemStream_Flag && process.locate!=LOCATE_INIT)
	{
		pts = AudioPTS/90000;
		sprintf(szBuffer, "%d:%02d:%02d", pts/3600, (pts%3600)/60, pts%60);
		SetDlgItemText(hDlg, IDC_TIMESTAMP, szBuffer);
	}
	else
		SetDlgItemText(hDlg, IDC_TIMESTAMP, "");

	sprintf(szBuffer,"%d / %d",(int)d2v_current.lba,d2v_current.file+StartVobNumber);
	SetDlgItemText(hDlg, IDC_SECTOR,szBuffer);

	if (process.locate==LOCATE_RIP)
	{
		if (NTSC_Purity || FILM_Purity)
		{
			if (!FILM_Purity)
			{
				if (frame_rate==25)
					sprintf(szBuffer, "PAL");
				else
					sprintf(szBuffer, "NTSC");
			}
			else if (!NTSC_Purity)
				sprintf(szBuffer, "FILM");
			else if (NTSC_Purity > Old_NTSC_Purity)
				sprintf(szBuffer, "NTSC %2d %%", 100 - (FILM_Purity*100)/(FILM_Purity+NTSC_Purity));
			else
				sprintf(szBuffer, "FILM %2d %%", (FILM_Purity*100)/(FILM_Purity+NTSC_Purity));

			Old_NTSC_Purity = NTSC_Purity;
			SetDlgItemText(hDlg, IDC_VIDEO_TYPE, szBuffer);
		}

		if (process.op)
		{
			process.ed = timeGetTime();
			process.elapsed = (process.ed-process.op)/1000;
			process.percent = (float)(100.0*(process.run-process.start+_telli64(Infile[File_Flag]))/(process.end-process.start));
			process.remain = (int)((process.ed-process.op)*(100.0-process.percent)/process.percent)/1000;

			sprintf(szBuffer, "%d:%02d:%02d", process.elapsed/3600, (process.elapsed%3600)/60, process.elapsed%60);
			SetDlgItemText(hDlg, IDC_ELAPSED, szBuffer);

			sprintf(szBuffer, "%d:%02d:%02d", process.remain/3600, (process.remain%3600)/60, process.remain%60);
			SetDlgItemText(hDlg, IDC_REMAIN, szBuffer);
		}
	}
	else if (GetDlgItemText(hDlg, IDC_ELAPSED, szBuffer, 9))
	{
		SetDlgItemText(hDlg, IDC_VIDEO_TYPE, "");
		SetDlgItemText(hDlg, IDC_FRAME_TYPE, "");
		SetDlgItemText(hDlg, IDC_CODED_NUMBER, "");
		SetDlgItemText(hDlg, IDC_PLAYBACK_NUMBER, "");
		SendDlgItemMessage(hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
		SetDlgItemText(hDlg, IDC_FILE, "");
		SetDlgItemText(hDlg, IDC_FILE_SIZE, "");
		SetDlgItemText(hDlg, IDC_ELAPSED, "");
		SetDlgItemText(hDlg, IDC_REMAIN, "");
		SetDlgItemText(hDlg, IDC_FPS, "");
	}
}
