/*
 *  This code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This code is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "MenuShrinkEngine.h"
#include <memory.h>
#include <stdlib.h>

// 

#define CHAR_SIZE 256

// This stuff is global and accessed by all the classes, for messages and errors.
// The string where messages are stored.
char Message[CHAR_SIZE];
// Our default function to handle regular and error messages.
void DefaultLogMessage(char* AMessage, int Langage)
{
	printf(AMessage);
}
// Global pointers that classes use to send regular and error messages.
void (*pMessageFunction)(char* AMessage, int Langage) = DefaultLogMessage;
void (*pErrorFunction)(char* AMessage, int Langage) = DefaultLogMessage;

int	 (*FrameSelectFunction)(int VobId, int CellId, int StartLBA, int EndLBA, int Default, 
							int NumLBA, int NormallySkipped);



// This is a bit ugly, but I'll live.
#define GROW_ARRAY(ARRAY,TYPE,OLDSIZE,NEWSIZE) {TYPE* Temp = new TYPE[NEWSIZE];\
memset(Temp,0,(NEWSIZE)*sizeof(TYPE));\
memcpy(Temp,ARRAY,(OLDSIZE)*sizeof(TYPE));\
delete[] ARRAY;\
ARRAY = Temp;}

// TODO: 
// - Verify offsets in IFO files to avoid exceptions.
// - Restore BUP files!
// - Add Get VTS sectors. (VMGI_MAT and VMG_PTT_SRPT)

// This utility splits OrigPath into a directory and file name.
void SplitPath(char* OrigPath, char* Dir, char* FileName, char* Ext)
{
	int i,j;
	// First, find extention:
	strcpy(Ext,"");
	for(i=strlen(OrigPath);i>=0;i--)
	{
		if(OrigPath[i] == '.')
		{
			// Found the extension Split right there.
			strcpy(Ext,OrigPath+i);
			break;
		}
	}
	j = (i==-1) ? strlen(OrigPath) : i;
	for(i=j;i>=0;i--)
	{
		if(OrigPath[i] == '/' || OrigPath[i] == '\\')
		{
			// Found the last / or \. Split right there.
			strncpy(Dir,OrigPath,i+1);
			Dir[i+1] = '\0';
			strncpy(FileName,OrigPath+i+1,j-i-1);
			FileName[j-i-1]='\0';
			return;
		}
	}
	// Didn't find / or \.
	strncpy(FileName,OrigPath,j);
	FileName[j] = '\0';
	strcpy(Dir,"");
}

#define BLOCK 256

CMenuShrinkEngine::CMenuShrinkEngine()
{
	m_Cell1 = NULL;
	m_Cell2 = NULL;
	m_CellPBInfo = new CELL_PLAYBACK_INFO[BLOCK];
	m_CellPosInfo = new CELL_POS_INFO[BLOCK];
	m_ADTInfo = new VTSM_C_ADTItem[BLOCK];
	m_CellXInfo = new CELL_EXTRA_INFO[BLOCK];

	m_NumAllocated = BLOCK;
	memset(m_CellPBInfo,0,m_NumAllocated*sizeof(CELL_PLAYBACK_INFO));
	memset(m_CellPosInfo,0,m_NumAllocated*sizeof(CELL_POS_INFO));
	memset(m_ADTInfo,0,m_NumAllocated*sizeof(VTSM_C_ADTItem));
	memset(m_CellXInfo,0,m_NumAllocated*sizeof(CELL_EXTRA_INFO));
	m_NumCells = 0;
	m_Busy = 0;
	m_FixSubPicStops = true;
	m_UseInfiniteStills = false;
	m_OneSecondStill = 0;
	m_SkipButtonlessCells = 1;
	m_MinButtonSize = 5;
	m_PreviewAllCells = 0;

	m_NumAllocNavPacks = BLOCK;
	m_NavPackLBATable = new uint[m_NumAllocNavPacks];
	m_NumNavPacks = 0;
	m_KeepAudio = 0;
	m_IFOFileSize = 0;
	m_AbortNow = 0;
	m_SectorsProcessedSoFar = 0;
	m_FinalCumulMenuSize = 0;
	m_OrigCumulMenuSize = 0;
	m_VTSCurrentlyProcessing = NULL;
	m_CurrentVTSIndex = 0;
	m_OverallDVDSize = 0;

	memset(m_VTSInfo,0,100*sizeof(VTSInfo));
	m_LastVTS = -1;
	m_FrameSelect = 85;
	FrameSelectFunction = NULL;

	strcpy(m_BackupDirName,"MenuShrinkBackup");
	strcpy(m_BackupDirFullPath,"");
	strcpy(m_WorkingDirFullPath,"");
	strcpy(m_BaseVTSName,"");

}


CMenuShrinkEngine::~CMenuShrinkEngine()
{
	delete m_Cell1;
	delete m_Cell2;
	delete[] m_CellPBInfo;
	delete[] m_CellPosInfo;
	delete[] m_ADTInfo;
	delete[] m_CellXInfo;
	delete[] m_NavPackLBATable;
}

int CheckAndCreateBackup(char* OriginalFileName, char* BackupFileName)
{
	// Open input file.
	FILE* InputFd = fopen(BackupFileName,"rb");
	if(InputFd == NULL)
	{
		// Try to create backup file by moving same file above.
		if(rename(OriginalFileName,BackupFileName) != 0)
		{
			ERROR2("I can't find file %s or I can't move it to the backup directory.\n",BackupFileName);
			ERRORF2("Fichier %s non trouv, ou impossible de le dplacer dans le rpertoire de sauvegarde.\n",BackupFileName);
			return -1;
		}
		InputFd = fopen(BackupFileName,"rb");
		if(InputFd == NULL)
		{
			ERROR2("Can't open backup file %s for reading\n",BackupFileName);
			ERRORF2("Impossible d'ouvrir le fichier %s en mode lecture\n",BackupFileName);
			return -1;
		}
	}
	fclose(InputFd);
	return 0;
}

int CMenuShrinkEngine::ProcessFullDVD()
{
	int i, ret;
	if(m_Busy)
		return -1;
	VideoFormat = FORMAT_UNKNOWN;
	m_NumSectorsToProcess = 0;
	m_SectorsProcessedSoFar = 0;

	MESSAGE3("Processing %s -- %s\n\n",m_WorkingDirFullPath, VERSION);
	MESSAGEF3("Traitement de %s -- %s\n\n",m_WorkingDirFullPath, VERSION);
	for(i=0;i<=m_LastVTS;i++)
	{
		if(m_VTSInfo[i].m_SkipThisVTS == 0)
			m_NumSectorsToProcess += m_VTSInfo[i].m_VTSMenuOrigSize;
	}
	m_Busy = 1;
	for(i=m_LastVTS;i>=0;i--)
	{
		// Do we want to process this one?
		if(m_VTSInfo[i].m_SkipThisVTS == 0 && m_VTSInfo[i].m_VTSMenuOrigSize != 0)
		{
			ret = ProcessVTS(i,m_KeepAudio,m_FrameSelect);
			if(ret == -1)
				goto end;
			m_VTSInfo[i].m_VTSMenuFinalSize = ret;
			if(m_AbortNow)
				break;
		}
		else if(i==0)
		{
			// We still need a backup of video_ts.ifo
			char BackupName[CHAR_SIZE];
			sprintf(BackupName,"%s/VIDEO_TS.IFO",m_BackupDirFullPath);
			if(CheckAndCreateBackup("VIDEO_TS.IFO",BackupName) == -1)
				goto end;
			sprintf(BackupName,"%s/VIDEO_TS.BUP",m_BackupDirFullPath);
			if(CheckAndCreateBackup("VIDEO_TS.BUP",BackupName) == -1)
				goto end;
		}

	}
	if(m_AbortNow)
	{
		ERROR1("Process aborted by user\n");
		ERRORF1("Traitement interrompu par l'utilisateur\n");
		m_Busy = 0;
		return -1;
	}
	FinalGetVTSSectors();
	// Compute final size:
	m_FinalCumulMenuSize = 0;
	for(i=0;i<=m_LastVTS;i++)
		m_FinalCumulMenuSize += m_VTSInfo[i].m_SkipThisVTS ? 
		m_VTSInfo[i].m_VTSMenuOrigSize : m_VTSInfo[i].m_VTSMenuFinalSize;
	// And compute final overall DVD size.
	GetOveralDVDSize();
	MESSAGE2("Done: Original overall menu size %.2fMB\n",2*(float)m_OrigCumulMenuSize/1024);
	MESSAGEF2("Fin du traitement: Taille originale des menus %.2fMB\n",2*(float)m_OrigCumulMenuSize/1024);
	MESSAGE3("Final overall menu size %.2fMB -- Savings: %.2fMB\n",
		2*(float)m_FinalCumulMenuSize/1024, 2*(float)m_OrigCumulMenuSize/1024-2*(float)m_FinalCumulMenuSize/1024);
	MESSAGEF3("Taille finale des menus %.2fMB -- Diffrence: %.2fMB\n",
		2*(float)m_FinalCumulMenuSize/1024, 2*(float)m_OrigCumulMenuSize/1024-2*(float)m_FinalCumulMenuSize/1024);
	
	m_Busy = 0;
	return 0;

end:
	m_Busy = 0;
	return -1;
}

// Helper function that moves a file from Src to Dest. If Src does not exist, nothing happens.
// If Dest exists, it is first removed.
int MoveBack(char* Src, char* Dest)
{
	FILE* fd = fopen(Src,"rb");
	if(fd != 0)
	{
		// Original exists. Let's remove destination and rename.
		fclose(fd);
		remove(Dest);
		if(rename(Src,Dest) != 0)
		{
			ERROR2("Couldn't restore file %s\n",Src);
			ERRORF2("Impossible de restaurer le fichier %s\n",Src);
			return -1;
		}
		return 0;
	}
	return 0;
}

int CMenuShrinkEngine::RestoreBackup()
{
	if(m_Busy)
		return -1;
	int Ret = 0;
	if(strcmp(m_BackupDirFullPath,"") != 0)
	{
		// Go to backup directory and try to move video_ts.ifo and .vob.
		if(_chdir(m_BackupDirFullPath) != 0)
		{
			MESSAGE1("Nothing to backup\n");
			MESSAGEF1("Pas de sauvegarde\n");
			return 0;
		}
		Ret |= MoveBack("VIDEO_TS.IFO","../VIDEO_TS.IFO");
		Ret |= MoveBack("VIDEO_TS.VOB","../VIDEO_TS.VOB");
		Ret |= MoveBack("VIDEO_TS.BUP","../VIDEO_TS.BUP");
		char Src[CHAR_SIZE];
		char Dest[CHAR_SIZE];
		int i;
		for(i=1;i<100;i++)
		{
			sprintf(Src,"VTS_%02d_0.IFO",i);
			sprintf(Dest,"../VTS_%02d_0.IFO",i);
			Ret |= MoveBack(Src,Dest);
			sprintf(Src,"VTS_%02d_0.VOB",i);
			sprintf(Dest,"../VTS_%02d_0.VOB",i);
			Ret |= MoveBack(Src,Dest);
			sprintf(Src,"VTS_%02d_0.BUP",i);
			sprintf(Dest,"../VTS_%02d_0.BUP",i);
			Ret |= MoveBack(Src,Dest);
		}
		// Don't forget to _chdir back where we were.
		_chdir("..");
		memset(m_VTSInfo,0,100*sizeof(VTSInfo));
		ScanDVDFolder(m_WorkingDirFullPath,false);
		if(Ret == 0)
		{
			MESSAGE1("Backup files restored\n");
			MESSAGEF1("Sauvegardes restaures\n");
		}
		GetOveralDVDSize();
		return 0;
	}
	return -1;
}

int CMenuShrinkEngine::ProcessVTS(int VTSNum,  int KeepAudio, FrameSelect WhichFrame)
{
	if(VTSNum < 0 || VTSNum > 99)
	{
		ERROR1("Vob number must be between 0 (for VMG) and 99\n");
		ERRORF1("Le numro de VOB doit tre entre 0 (pour le VMG) et 99\n");
		return -1;
	}
	if(VTSNum > 0)
		sprintf(m_InFileName,"VTS_%02d_0.VOB",VTSNum);
	else
		sprintf(m_InFileName,"VIDEO_TS.VOB");

	m_CurrentVTSIndex = VTSNum;
	return ProcessVTS(m_InFileName,KeepAudio,WhichFrame);
}

int CMenuShrinkEngine::ProcessVTS(char* VTSFileName, int KeepAudio, FrameSelect WhichFrame)
{
	// See if the backup directory exists.
	char dir[CHAR_SIZE];
	char fname[CHAR_SIZE];
	char ext[16];
	char InputFileName[CHAR_SIZE];
	char IFOInputFileName[CHAR_SIZE];
	char IFOOutputFileName[CHAR_SIZE];
	char BUPInputFileName[CHAR_SIZE];
	char BUPOutputFileName[CHAR_SIZE];

	// Split path.
	SplitPath( VTSFileName, dir, fname, ext);
	// Create backup directory name.
	sprintf(m_BackupDirFullPath,"%s%s",dir,m_BackupDirName);
	// Try to _chdir into it. 
	if(_chdir(m_BackupDirFullPath) != 0)
	{
		// Directory does not exist, or isn't accessible? Try to create it.
		if(mkdir(m_BackupDirFullPath) != 0 || _chdir(m_BackupDirFullPath) != 0)
		{
			char Temp[1024];
			_getcwd(Temp,1024);
			ERROR2("Current directory is %s\n",Temp);
			ERRORF2("Rpertoire courant: %s\n",Temp);
			
			ERROR2("Can't access or create backup folder %s\n",m_BackupDirFullPath);
			ERRORF2("Impossible de trouver ou de crer le rpertoire de sauvegarde %s\n",m_BackupDirFullPath);
			return -1;
		}
	}
	_chdir("..");
	// See if backup file exists. If not create it!
	sprintf(InputFileName,"%s/%s.VOB",m_BackupDirFullPath,fname);
	if(CheckAndCreateBackup(VTSFileName,InputFileName) == -1)
		return -1;
	// Same for IFO.
	sprintf(IFOInputFileName,"%s/%s.IFO",m_BackupDirFullPath,fname);
	sprintf(IFOOutputFileName,"%s%s.IFO",dir,fname);
	if(CheckAndCreateBackup(IFOOutputFileName,IFOInputFileName) == -1)
		return -1;
	// Same for BUP.
	sprintf(BUPInputFileName,"%s/%s.BUP",m_BackupDirFullPath,fname);
	sprintf(BUPOutputFileName,"%s%s.BUP",dir,fname);
	CheckAndCreateBackup(BUPOutputFileName,BUPInputFileName);

	// Create base VTS file name.
	if(strcmp(fname,"VIDEO_TS") == 0)
		strcpy(m_BaseVTSName,"VIDEO_TS");
	else
	{
		strncpy(m_BaseVTSName,fname,strlen(fname)-1);
		m_BaseVTSName[strlen(fname)-1]='\0';
	}

	// We're good to go!
	if(WhichFrame == 9999)
	{
		MESSAGE3("Shrinking %s,%s keeping audio,\nselecting the largest I-frame.\n",VTSFileName,
			(KeepAudio)?"":" not");
		MESSAGEF3("Traitement de %s,%s son,\nslection de la plus grande image.\n",VTSFileName,
			(KeepAudio)?" avec":" sans");
	}
	else
	{
		MESSAGE3("Shrinking %s,%s keeping audio\n",VTSFileName, (KeepAudio)?"":" not");
		MESSAGEF3("Traitement de %s,%s son.\n",VTSFileName,(KeepAudio)?" avec":" sans");
		MESSAGE2("Selecting the I-frame %d%% into the menu.\n",WhichFrame);
		MESSAGEF2("Slection de l'image situe  %d%% de la dure du menu.\n",WhichFrame);
	}

	ResetAll();
	m_VTSCurrentlyProcessing = InputFileName;
	int EndSize = ShrinkAllCells(InputFileName,VTSFileName,KeepAudio, WhichFrame);
	int ret = 0;
	if(EndSize != -1 && m_AbortNow != 1)
	{
		// Create IFO name:
		char IFOOutputName[CHAR_SIZE];
		sprintf(IFOOutputName,"%s%s.IFO",dir,fname);
		ret = UpdateIFO(IFOInputFileName,IFOOutputName, strcmp(fname,"VIDEO_TS") == 0);
	}
	m_VTSCurrentlyProcessing = NULL;
	if(ret == 0 && EndSize != -1)
	{
		MESSAGE1("Done...\n\n\n");
		MESSAGEF1("Traitement fini...\n\n\n");
	}
	else
	{
		MESSAGE1("Done with errors!!\n\n\n");
		MESSAGEF1("Traitement fini avec erreur!!\n\n\n");
	}
	return EndSize;
}

void CMenuShrinkEngine::ResetAll()
{
	delete m_Cell1;
	delete m_Cell2;
	delete[] m_CellPBInfo;
	delete[] m_CellPosInfo;
	delete[] m_ADTInfo;
	delete[] m_CellXInfo;
	delete[] m_NavPackLBATable;

	m_Cell1 = NULL;
	m_Cell2 = NULL;
	m_CellPBInfo = new CELL_PLAYBACK_INFO[BLOCK];
	m_CellPosInfo = new CELL_POS_INFO[BLOCK];
	m_ADTInfo = new VTSM_C_ADTItem[BLOCK];
	m_CellXInfo = new CELL_EXTRA_INFO[BLOCK];

	m_NumAllocated = BLOCK;
	memset(m_CellPBInfo,0,m_NumAllocated*sizeof(CELL_PLAYBACK_INFO));
	memset(m_CellPosInfo,0,m_NumAllocated*sizeof(CELL_POS_INFO));
	memset(m_ADTInfo,0,m_NumAllocated*sizeof(VTSM_C_ADTItem));
	memset(m_CellXInfo,0,m_NumAllocated*sizeof(CELL_EXTRA_INFO));
	m_NumCells = 0;

	m_NumAllocNavPacks = BLOCK;
	m_NavPackLBATable = new uint[m_NumAllocNavPacks];
	m_NumNavPacks = 0;
	m_AbortNow = 0;
}

int CMenuShrinkEngine::GatherCellInfo(CCell* TheCell)
{
	int i;
	// Grab everything we'll need to update the IFO.
	if(m_NumCells == m_NumAllocated)
	{
		// Running out of space. Let's allocate more...

		GROW_ARRAY(m_CellPBInfo,CELL_PLAYBACK_INFO,m_NumAllocated,m_NumAllocated+BLOCK);
		GROW_ARRAY(m_CellPosInfo,CELL_POS_INFO,m_NumAllocated,m_NumAllocated+BLOCK);
		GROW_ARRAY(m_ADTInfo,VTSM_C_ADTItem,m_NumAllocated,m_NumAllocated+BLOCK);
		GROW_ARRAY(m_CellXInfo,CELL_EXTRA_INFO,m_NumAllocated,m_NumAllocated+BLOCK);

		m_NumAllocated += BLOCK;
	}

	// Check for interleaved VOBUs. We would detect that if the VOB/CELL ID of this cell was
	// already in our list. Not too sure what we should do in such cases, but exiting for now.
	for(i=0;i<m_NumCells;i++)
	{
		if(m_CellPosInfo[i].m_VOBID == TheCell->m_VOBID && m_CellPosInfo[i].m_CELLID == TheCell->m_CELLID)
		{
			// The cell already is in our list! ARGH! This means the VOB is interleaved. We have to exit.
			ERROR1("FATAL PROBLEM: the VOB file appears to have interleaved VOBUs. Can't process such files\n");
			ERRORF1("IMPOSSIBLE DE CONTINUER: Le fichier VOB semble avoir des VOBUs entrelaces.\n");
			return -1;
		}
	}

	// This is for the cel list in the PGC info.
	m_CellPBInfo[m_NumCells].m_FirstVOBUStartLBA = TheCell->m_StartLBA;
	int LastNav = TheCell->FindPrevNavPack(TheCell->m_NumPacks);
	m_CellPBInfo[m_NumCells].m_LastVOBUStartLBA = LastNav+TheCell->m_StartLBA;
	m_CellPBInfo[m_NumCells].m_LastVOBUEndLBA = TheCell->m_StartLBA + TheCell->m_NumPacks-1;
	if(!TheCell->m_KeepAudio && TheCell->m_FoundButtons)
	{
		// If we're not keeping the audio and we found buttons, let's turn that into a still.
		m_CellPBInfo[m_NumCells].m_Still = 255;
	}
	// Also, if we're not keeping audio, the cell has only 1 frame.
	if(!TheCell->m_KeepAudio)
		m_CellPBInfo[m_NumCells].m_PBTime = 1;

	if(!TheCell->m_FoundButtons)
	{
		// Kill this cell if it does not have buttons.
		m_CellXInfo[m_NumCells].m_KillIt = 1;
	}
	// TODO JEAN: UPDATE TIME if I killed that cell.
	m_CellPosInfo[m_NumCells].m_VOBID = TheCell->m_VOBID;
	m_CellPosInfo[m_NumCells].m_CELLID = TheCell->m_CELLID;

	m_ADTInfo[m_NumCells].m_VOBID = TheCell->m_VOBID;
	m_ADTInfo[m_NumCells].m_CELLID = TheCell->m_CELLID;
	m_ADTInfo[m_NumCells].m_StartLBA = TheCell->m_StartLBA;
	m_ADTInfo[m_NumCells].m_EndLBA = TheCell->m_StartLBA + TheCell->m_NumPacks-1;

	m_CellXInfo[m_NumCells].m_KeepAudio = TheCell->m_KeepAudio;
	m_CellXInfo[m_NumCells].m_AnyLinkNext = TheCell->m_AnyLinkNext;
	m_CellXInfo[m_NumCells].m_KeepAnimated = 0;
	m_CellXInfo[m_NumCells].m_CellDuration90 = TheCell->m_EndPTS - TheCell->m_StartPTS;
	m_CellXInfo[m_NumCells].m_NumAudioStreams = TheCell->m_NumFoundAudioStreams;
	memcpy(m_CellXInfo[m_NumCells].m_AudioStreams,TheCell->m_FoundAudioStreams,8*sizeof(uint));
	
	m_NumCells++;

	// Now for the nav pack table.
	for(i=0;i<TheCell->m_NumPacks;i++)
	{
		// Go through the packs in the cell, if you find a navpack, keep track of its LBA.
		if(TheCell->m_Packs[i].m_PackType == PackTypeNavPack)
		{
			// Grow table if needed.
			if(m_NumNavPacks == m_NumAllocNavPacks)
			{
				GROW_ARRAY(m_NavPackLBATable,uint,m_NumAllocNavPacks,m_NumAllocNavPacks+BLOCK);
				m_NumAllocNavPacks += BLOCK;
			}
			m_NavPackLBATable[m_NumNavPacks++] = TheCell->m_StartLBA+i;
		}
	}
	return 0;
}

int CMenuShrinkEngine::ShrinkAllCells(char* InFileName, char* OutFileName, int KeepAudio, FrameSelect WhichFrame)
{
	delete m_Cell1;
	delete m_Cell2;
	m_Cell2 = NULL;
	m_Cell1 = NULL;
	int ret = 0;
	int LastLBA = 0;
	m_KeepAudio = KeepAudio;

	FILE* Out = fopen(OutFileName,"w+b");
	if(Out == NULL)
	{
		ERROR2("Can't open %s for writing, exiting\n",OutFileName);
		ERRORF2("Impossible d'ouvrir %s en mode criture\n",OutFileName);
		return -1;
	}

	FILE* In = fopen(InFileName,"rb");
	if(In == NULL)
	{
		ERROR2("Can't open %s for reading, exiting\n",InFileName);
		ERRORF2("Impossible d'ouvrir %s en mode lecture\n",InFileName);
		return -1;
	}
	m_Cell1 = new CCell(In);
	m_Cell1->m_MinButtonSize = m_MinButtonSize;
	m_Cell1->m_PreviewAllCells = m_PreviewAllCells;

	// Find the cell in question.
	ret = m_Cell1->ReadFromFile(In, m_SectorsProcessedSoFar);
	MESSAGE5("\nFound VOB/CELL ID %d/%d with %d buttons at sector %d\n", 
		m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
//if(m_Cell1->m_Packs[0].m_NumButtons > 0)
//MESSAGE3("Max button size %dx%d\n",m_Cell1->m_MaxSizeX,m_Cell1->m_MaxSizeY);
	MESSAGEF5("\nCellule VOB/CELL ID %d/%d trouve avec %d boutons au secteur %d\n", 
		m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
int i=0;
	while(ret == 0 && m_AbortNow == 0)
	{
		int Iframe = WhichFrame;
		// Create output Cell.
		delete m_Cell2; m_Cell2 = NULL;
		m_Cell2 = new CCell(m_Cell1->m_NumPacks);
		m_Cell2->m_MinButtonSize = m_MinButtonSize;
		m_Cell2->m_PreviewAllCells = m_PreviewAllCells;
		m_Cell2->m_FixSubPicStops = m_FixSubPicStops;
		// This does most of the work...
		ret = m_Cell2->CopySingleIFrame(m_Cell1,Iframe,KeepAudio!=0);
		if(ret != -2)
		{
			// Then write back to file.
			ret = m_Cell2->WriteToFile(In,Out);
			// Get all the info we need.
			ret = GatherCellInfo(m_Cell2);
		}
		else
		{
			// Here, the user decided to keep the cell animated.
			ret = m_Cell1->WriteToFile(In,Out,0);
			// Get all the info we need.
			m_Cell1->m_KeepAudio = 1;
			m_Cell1->m_FoundButtons = 1;
			ret = GatherCellInfo(m_Cell1);
			m_CellXInfo[m_NumCells-1].m_KeepAnimated = 1;
			m_CellXInfo[m_NumCells-1].m_KillIt = 0;
		}
		if(ret != 0)
		{
			// Interleaved VOBUs problem. let's return -1.
			return -1;
		}
		// On to the next cell.
		if(m_AbortNow)
			break;
		ret = m_Cell1->ReadNextCellFromFile(In, m_SectorsProcessedSoFar);
		if(ret == 0)
		{
			MESSAGE5("\nFound VOB/CELL ID %d/%d with %d buttons at sector %d\n", 
				m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
//if(m_Cell1->m_Packs[0].m_NumButtons > 0)
//MESSAGE3("Max button size %dx%d\n",m_Cell1->m_MaxSizeX,m_Cell1->m_MaxSizeY);
			MESSAGEF5("\nCellule VOB/CELL ID %d/%d trouve avec %d boutons au secteur %d\n", 
				m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
		}
#if 0
i++;
if(i==2)
break;
#else
#endif
	}
	int OutLBACount = ftell(Out) / 1024 / 2;
	fclose(Out);
	fclose(In);
	return OutLBACount;
}

int CMenuShrinkEngine::ShrinkOneCell(char* InFileName, char* OutFileName, ushort TheVOBID, uchar TheCELLID, int KeepAudio, 
									 FrameSelect WhichFrame)
{
	FILE* Out = 0;
	delete m_Cell1;
	delete m_Cell2;
	m_Cell2 = NULL;
	m_Cell1 = NULL;
	int ret = 0;
	int LastLBA = 0;
	m_KeepAudio = KeepAudio;

	FILE* In = fopen(InFileName,"rb");
	if(In == NULL)
	{
		ERROR2("Can't open %s for reading, exiting\n",InFileName);
		ERRORF2("Impossible d'ouvrir %s en mode lecture\n",InFileName);
		return -1;
	}
	m_Cell1 = new CCell(In);
	m_Cell1->m_MinButtonSize = m_MinButtonSize;
	m_Cell1->m_PreviewAllCells = m_PreviewAllCells;

	// Find the cell in question.
	ret = m_Cell1->FindThisCell(In,TheVOBID,TheCELLID);
	if(ret != -1)
	{
		// Load cell into input cell.
		m_Cell1->ReadFromFile(In, m_SectorsProcessedSoFar);
		MESSAGE5("Found VOB/CELL ID %d/%d with %d buttons at sector %d\n", 
			m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
		MESSAGEF5("Cellule VOB/CELL ID %d/%d trouve avec %d boutons au sectur %d\n", 
			m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
		// Open output file for writing.
		Out = fopen(OutFileName,"w+b");
		if(Out == NULL)
		{
			ERROR2("Can't open %s for writing, exiting\n",OutFileName);
			ERRORF2("Impossible d'ouvrir %s en mode criture\n",OutFileName);
			return 0;
		}
		fclose(Out);
		Out = fopen(OutFileName,"r+b");
		int Iframe = WhichFrame;
		m_Cell2 = new CCell(m_Cell1->m_NumPacks);
		m_Cell2->m_MinButtonSize = m_MinButtonSize;
		m_Cell2->m_PreviewAllCells = m_PreviewAllCells;
		m_Cell2->m_FixSubPicStops = m_FixSubPicStops;
		ret = m_Cell2->CopySingleIFrame(m_Cell1,Iframe,KeepAudio!=0);
		if(ret == -2)
			ret = m_Cell1->WriteToFile(In,Out,0);	// The user decided to keep this cell intact!
		else
			ret = m_Cell2->WriteToFile(In,Out);
	}
	else
	{
		// Couldn't find cell.
		MESSAGE3("Reached end of file while looking for VOB/CELL ID %d/%d\n",
			TheVOBID,TheCELLID);
		MESSAGEF3("Fin du fichier, la cellule VOB/CELL ID %d/%d n'a pas t trouve\n",
			TheVOBID,TheCELLID);
		return 0;
	}
	int OutLBACount = 0;
	if(Out)
	{
		OutLBACount = ftell(Out) / 1024 / 2;
		fclose(Out);
	}
	return OutLBACount;
}

int CMenuShrinkEngine::ExtractOneCell(char* InFileName, char* OutFileName, ushort TheVOBID, uchar TheCELLID)
{
	FILE* Out = 0;
	delete m_Cell1;
	m_Cell1 = NULL;
	int ret = 0;
	int LastLBA = 0;

	FILE* In = fopen(InFileName,"rb");
	if(In == NULL)
	{
		ERROR2("Can't open %s for reading, exiting\n",InFileName);
		ERRORF2("Impossible d'ouvrir %s en mode lecture\n",InFileName);
		return -1;
	}
	m_Cell1 = new CCell(In);
	m_Cell1->m_MinButtonSize = m_MinButtonSize;
	m_Cell1->m_PreviewAllCells = m_PreviewAllCells;

	// Find the cell in question.
	ret = m_Cell1->FindThisCell(In,TheVOBID,TheCELLID);
	if(ret != -1)
	{
		// Load cell into input cell.
		m_Cell1->ReadFromFile(In, m_SectorsProcessedSoFar);
		MESSAGE5("Found VOB/CELL ID %d/%d with %d buttons at sector %d\n", 
			m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
		MESSAGEF5("Cellule VOB/CELL ID %d/%d avec %d boutons au secteur %d\n", 
			m_Cell1->m_VOBID,m_Cell1->m_CELLID,m_Cell1->m_Packs[0].m_NumButtons,m_Cell1->m_StartLBA);
		// Open output file for writing.
		Out = fopen(OutFileName,"wb");
		if(Out == NULL)
		{
			ERROR2("Can't open %s for writing, exiting\n",OutFileName);
			ERRORF2("Impossible d'ouvrir %s en mode criture\n",OutFileName);
			return 0;
		}
		ret = m_Cell1->WriteToFile(In,Out,0);
	}
	else
	{
		// Couldn't find cell.
		MESSAGE3("Reached end of file while looking for VOB/CELL ID %d/%d\n",
			TheVOBID,TheCELLID);
		MESSAGEF3("Fin du fichier, la cellule VOB/CELL ID %d/%d n'a pas t trouve\n",
			TheVOBID,TheCELLID);
		return 0;
	}
	int OutLBACount = 0;
	if(Out)
	{
		OutLBACount = ftell(Out) / 1024 / 2;
		fclose(Out);
	}
	return OutLBACount;
}

#define MAX_IFO_GROWTH_LBA	40	// Maximum number of sectors we'll let the IFO grow.

int CMenuShrinkEngine::UpdateIFO(char* IFONameInput, char* IFONameOutput, bool VMG)
{
	// open input and output IFO files.
	FILE* fdIn = fopen(IFONameInput,"rb");
	if(fdIn == NULL)
	{
		ERROR2("Could not open %s for update, exiting\n",IFONameInput);
		ERRORF2("Impossible de modifier %s\n",IFONameInput);
		return -1;
	}
	FILE* fdOut = fopen(IFONameOutput,"wb");
	if(fdOut == NULL)
	{
		ERROR2("Could not open %s for update, exiting\n",IFONameOutput);
		ERRORF2("Impossible de modifier %s\n",IFONameOutput);
		return -1;
	}
	MESSAGE2("Updating %s\n",IFONameOutput);
	MESSAGEF2("Traitement de %s\n",IFONameOutput);
	// Load input IFO file.
	fseek(fdIn,0,SEEK_END);
	m_IFOFileSize = ftell(fdIn);
	uchar* IFOFile = new uchar[m_IFOFileSize + MAX_IFO_GROWTH_LBA*SECTOR_SIZE];
	memset(IFOFile,0,m_IFOFileSize + MAX_IFO_GROWTH_LBA*SECTOR_SIZE);
	fseek(fdIn,0,SEEK_SET);
	fread(IFOFile,1,m_IFOFileSize,fdIn);
	fclose(fdIn);

	// Now correct all the IFO pointers and data.
	int DeltaLBA = 0;
	int OrigEndLBA = 0;
	int ret = 0;
	ret = CorrectPGCInfos(IFOFile,m_NumCells,m_CellPBInfo,m_CellPosInfo,VMG,m_CellXInfo);
	if(ret != 0) goto end;
	ret = CorrectVTSM_C_ADT(IFOFile,m_NumCells,m_ADTInfo);
	if(ret != 0) goto end;
	ret = CorrectVTSM_VOBU_ADMAP(IFOFile, m_NumNavPacks, m_NavPackLBATable, OrigEndLBA, DeltaLBA);
	if(ret != 0) goto end;
	ret = CorrectLocalVTSSectors(IFOFile, OrigEndLBA, DeltaLBA, VMG);
	if(ret != 0) goto end;

	// Write back to output IFO.
	ret = fwrite(IFOFile,1,m_IFOFileSize,fdOut);
	if(ret != m_IFOFileSize)
	{
		ERROR2("Couldn't write IFO file %s\n",IFONameOutput);
		ERRORF2("Impossible de sauver le fichier IFO %s\n",IFONameOutput);
		ret = -1;
	}
	else
		ret = 0;
	fclose(fdOut);
	// Copy BUP file too! This is incredibly low tech. I can't find a copy() function to copy one file
	// to another.
	char BUPName[CHAR_SIZE];
	strcpy(BUPName,IFONameOutput);
	int i;
	for(i=strlen(BUPName)-1;i>0;i--)
	{
		if(BUPName[i] == '.')
			break;
	}
	strcpy(BUPName+i,".BUP");
	fdOut = fopen(BUPName,"wb");
	if(fdOut == NULL)
	{
		ERROR2("Could not open %s for update, exiting\n",BUPName);
		ERRORF2("Impossible de modifier %s\n",BUPName);
		delete[] IFOFile;
		return -1;
	}
	fwrite(IFOFile,1,m_IFOFileSize,fdOut);
	fclose(fdOut);
end:
	delete[] IFOFile;
	return ret;
}

int CMenuShrinkEngine::CorrectPGCInfos(uchar* Buffer, uint NumEntries, CELL_PLAYBACK_INFO* VOBCellPBInfo, 
									   CELL_POS_INFO* VOBCellPosInfo, bool VMG, CELL_EXTRA_INFO*	CellXInfo)
{
	IFO_main*	MainIFO;
	uint		i,j,m;
	int k;

	// Load main IFO structure.
	MainIFO = (IFO_main*)Buffer;
	// Find pointer to VTSM_PGCI_UT
	uint m_PVTSM_PGCI_UT;
	if(!VMG)
		m_PVTSM_PGCI_UT = GetInt(Buffer,0x00d0);
	else
		m_PVTSM_PGCI_UT = GetInt(Buffer,0x00c8);
	if(m_PVTSM_PGCI_UT == 0)
		return 0;
	// Read number of Language Units
	uint NumLU = GetShort(Buffer,m_PVTSM_PGCI_UT*SECTOR_SIZE);
	for(i=0;i<NumLU;i++)
	{
		// Find offset to VTSM_LU
		uint PVTSM_LU = GetInt(Buffer,m_PVTSM_PGCI_UT*SECTOR_SIZE+12+i*8);
		PVTSM_LU += m_PVTSM_PGCI_UT*SECTOR_SIZE;
		// Go there and get the number of PGCs.
		uint NumPGC = GetShort(Buffer,PVTSM_LU);
		for(j=0;j<NumPGC;j++)
		{
			// Find offset to VTSM_PGC
			uint PVTSM_PGC = GetInt(Buffer,PVTSM_LU+12+j*8);
			PVTSM_PGC += PVTSM_LU;
			bool PGCHasAudio = 0;
			// Find number of cells;
			uint NumCells = Buffer[PVTSM_PGC+3];
			// Find playback info.
			uint BasePCELL_PLAYBACK_INFO = PVTSM_PGC + GetShort(Buffer,PVTSM_PGC+0xe8);
			CELL_PLAYBACK_INFO* PGCCellPBInfo = (CELL_PLAYBACK_INFO*)(Buffer+BasePCELL_PLAYBACK_INFO);
			// And cell position info
			uint BasePCELL_POS_INFO = PVTSM_PGC + GetShort(Buffer,PVTSM_PGC+0xea);
			CELL_POS_INFO* PGCCellPosInfo = (CELL_POS_INFO*)(Buffer+BasePCELL_POS_INFO);

			CDVDTime TotalPGCTime;
			ushort VOBID;
			// It's important to go from last to first in case we skip cells. That's because we'll copy
			// k+1 into k. Going from top to 0 ensures that k+1 will already have the right stuff.
			for(k=(int)NumCells-1;k>=0;k--)
			{
				// Go through our list, see if we have a corresponding VOB/CELL ID and correct LBAs.
				for(m=0;m<NumEntries;m++)
				{
					VOBID = GetShort((uchar*)(&PGCCellPosInfo[k].m_VOBID),0);
					if(VOBID == VOBCellPosInfo[m].m_VOBID && PGCCellPosInfo[k].m_CELLID == VOBCellPosInfo[m].m_CELLID)
						break;
				}
				if(m<NumEntries)
				{
					// Found a match. This is just to remember that this cell was referenced.
					CellXInfo[m].m_WasReferenced = 1;
#if 0
					if(CellXInfo[m].m_KeepAudio != 0)
						PGCHasAudio = true;
#else
					if(CellXInfo[m].m_NumAudioStreams != 0)
						PGCHasAudio = true;
#endif
					if(m_SkipButtonlessCells)
					{
						// If there's a cell after us, and this one is to be blanked (m_BTTime == 1)
						// and does not have a cell command, we know for sure the next cell to play is
						// the next one in the PGC. So how about copying the next cell info into this
						// one? 
						// Here, I check whether the cell after us (which I'm going to use to replace this one
						// has any LinkNextCell or LinkNextPG command. If it does, I can't use it to replace
						// this one, because it will then repeat if the button is pressed.
						int NextHasLink = 0;
						if(k < (int)NumCells - 1 && CellXInfo[m].m_KillIt && PGCCellPBInfo[k].m_Command == 0)
						{
							// Get the VCID for the next cell in the PGC.
							ushort VID = GetShort((uchar*)(&PGCCellPosInfo[k+1].m_VOBID),0);
							uchar CID = PGCCellPosInfo[k+1].m_CELLID;
							// Now, find the corresponding cell in our own list.
							uint u;
							for(u=0;u<NumEntries;u++)
							{
								if(VID == VOBCellPosInfo[u].m_VOBID && CID == VOBCellPosInfo[u].m_CELLID)
								{
									// Found it! Check whether this cell has any LinkNextCell or LinkNextPG.
									NextHasLink = CellXInfo[u].m_AnyLinkNext;
									break;
								}
							}
							// Just to alert me to this fact!
							assert(NextHasLink == 0);
							if(NextHasLink)
							{
								MESSAGE4("NOTE: Can't skip LU %d PGC %d cell %d. Next one has link commands\n",i+1,j+1,k+1);
								MESSAGEF4("NOTE: Impossible d'effacer LU %d PGC %d cellule %d. Il y a des commandes Link dans la suivante.\n",i+1,j+1,k+1);
							}
						}

						if(!NextHasLink && k < (int)NumCells - 1 && CellXInfo[m].m_KillIt && PGCCellPBInfo[k].m_Command == 0)
						{
							MESSAGE5("Skipping LU %d PGC %d cell %d/%d\n",i+1,j+1,
								VOBID,PGCCellPosInfo[k].m_CELLID);
							MESSAGEF5("Effacement de la cellule: LU %d PGC %d cellule %d/%d\n",i+1,j+1,
								VOBID,PGCCellPosInfo[k].m_CELLID);
							PGCCellPBInfo[k] = PGCCellPBInfo[k+1];
							PGCCellPosInfo[k] = PGCCellPosInfo[k+1];
							goto AddTimes; // Ugly goto, but the alternative is even uglier!
						}
					}
					// Update LBA pointers.
					PutInt((uchar*)(&PGCCellPBInfo[k].m_FirstVOBUStartLBA),0,VOBCellPBInfo[m].m_FirstVOBUStartLBA);
					PutInt((uchar*)(&PGCCellPBInfo[k].m_LastVOBUStartLBA),0,VOBCellPBInfo[m].m_LastVOBUStartLBA);
					PutInt((uchar*)(&PGCCellPBInfo[k].m_LastVOBUEndLBA),0,VOBCellPBInfo[m].m_LastVOBUEndLBA);
					// Update still time, but only if the original was not a still.
					if(VOBCellPBInfo[m].m_Still != 0 && PGCCellPBInfo[k].m_Still == 0)
					{
						if(!m_UseInfiniteStills)
						{
							// Let's put a still equal to the original duration
							uint PBTime = GetInt((uchar*)(&PGCCellPBInfo[k].m_PBTime),0);
							int min = (PBTime & 0x00FF0000)>>16; min = min%16+10*(min/16);
							int sec = (PBTime & 0x0000FF00)>>8; sec = sec%16+10*(sec/16);
							sec += min*60;
							if(sec>60) sec = 60;
							if(sec<5) sec = 5;
							PGCCellPBInfo[k].m_Still = sec;
						}
						else
						{
							PGCCellPBInfo[k].m_Still = 255;
						}
					}
					else
					{
//						if(VOBCellPBInfo[m].m_Still == 0 && VOBCellPBInfo[m].m_PBTime == 1)
						if(CellXInfo[m].m_KillIt)
						{
							// This is a buttonless cell with only 1 frame. 
							// If needed, let's put a still of 1sec to avoid problems.
							if(m_OneSecondStill == 1)
								PGCCellPBInfo[k].m_Still = 1;
							// Otherwise let's put a still of 0! Useful for FBI warnings in their own PGCs.
							else
								PGCCellPBInfo[k].m_Still = 0;
						}
					}
					if(VOBCellPBInfo[m].m_PBTime == 1)
					{
						// The cell only has 1 frame, let's adjust its playback time. (12 frames for PAL, 18 for NTSC)
						if(VideoFormat == FORMAT_PAL)
							PGCCellPBInfo[k].m_PBTime = 0x52000000;
						else
							PGCCellPBInfo[k].m_PBTime = 0xd8000000;
					}
					else
					{
						// Normally, this is not needed, as the duration of the shouldn't change (we
						// kept the audio). But for reauthoring purposes (IFO out of sync) this can be
						// good.
						CDVDTime CellDur(CellXInfo[m].m_CellDuration90);
						CellDur.WriteToBuffer((uchar*) &PGCCellPBInfo[k].m_PBTime);
					}
					// Update flags: set SRC discontinuity and seamless playback of the next cell.
					PGCCellPBInfo[k].m_Flags &= 0x08;	// reset seamless playback.
					PGCCellPBInfo[k].m_Flags &= 0x01;	// reset seamless angle. CRUCIAL!!!
					PGCCellPBInfo[k].m_Flags |= 0x02;	// set STC(SRC) discontinuity.
					if(k<(int)NumCells-1)
					{
						PGCCellPBInfo[k+1].m_Flags &= 0x08;	// reset seamless playback.
						PGCCellPBInfo[k+1].m_Flags &= 0x01;	// reset seamless angle. CRUCIAL!!!
						PGCCellPBInfo[k+1].m_Flags |= 0x02;	// set STC(SRC) discontinuity.
					}
				}
				else
				{
					// Didn't find a match! This should never happen, right?
					MESSAGE5("NOTE: LU %d PGC %d uses a cell (%d/%d) that was not found in the VOB!\n",i+1,j+1,VOBID,PGCCellPosInfo[k].m_CELLID);
					MESSAGEF5("NOTE: LU %d PGC %d utilise une cellule (%d/%d) absente dans le fichier VOB!\n",i+1,j+1,VOBID,PGCCellPosInfo[k].m_CELLID);
				}
AddTimes:
				// Add cell playback time to TotalPGCTime.
				CDVDTime CellPBTime;
				CellPBTime.ReadFromBuffer((uchar*)(&PGCCellPBInfo[k].m_PBTime));
				if(CellPBTime.m_format != VideoFormat)
				{
					MESSAGE4("NOTE: LU %d PGC %d Cell %d has a cell time indicating the wrong video format!\n",i+1,j+1,m);
					MESSAGEF4("NOTE: LU %d PGC %d Cellule %d: l'info de dure indique le mauvais format vido!\n",i+1,j+1,m);
				}
				TotalPGCTime += CellPBTime;
			}
			// Set audio stream status for that PGC.
			if(!PGCHasAudio)
				memset(Buffer+PVTSM_PGC+0x0c,0,16);
			// Write back total PGC time.
			TotalPGCTime.WriteToBuffer(Buffer+PVTSM_PGC+4);
		}
	}
	
	// Let's just warn the user of any cell that was not originally referenced.
	for(i=0;i<NumEntries;i++)
	{
		if(CellXInfo[i].m_WasReferenced == 0)
		{
			// This cell was unreferenced!
			MESSAGE3("NOTE: Cell %d/%d was unreferenced in the original DVD\n",VOBCellPosInfo[i].m_VOBID,VOBCellPosInfo[i].m_CELLID);
			MESSAGEF3("NOTE: La cellule %d/%d n'est pas rfrence dans le DVD original\n",VOBCellPosInfo[i].m_VOBID,VOBCellPosInfo[i].m_CELLID);
		}
	}
	return 0;
}

int CMenuShrinkEngine::CorrectVTSM_C_ADT(uchar* Buffer, int NumEntries, VTSM_C_ADTItem* VOBCellADT)
{
	IFO_main*	MainIFO;
	uint		i;

	MainIFO = (IFO_main*)Buffer;
	
	// Find pointer to m_PVTSM_C_ADT
	uint m_PVTSM_C_ADT = GetInt((uchar*)&MainIFO->m_PVTSM_C_ADT,0);
	if(m_PVTSM_C_ADT == 0)
		return 0;

	// Read end byte, find number of entries.
	uint EndByte = GetInt(Buffer,m_PVTSM_C_ADT*SECTOR_SIZE+4);
	uint OrigNumEntries = (EndByte+1-8)/sizeof(VTSM_C_ADTItem);

	// Load original values.
	VTSM_C_ADTItem* PGCCellADT = (VTSM_C_ADTItem*)(Buffer+m_PVTSM_C_ADT*SECTOR_SIZE+8);
	// Endian conversion.
	for(i=0;i<OrigNumEntries;i++)
	{
		PGCCellADT[i].m_VOBID = GetShort((uchar*)(&PGCCellADT[i].m_VOBID),0);
		PGCCellADT[i].m_StartLBA = GetInt((uchar*)(&PGCCellADT[i].m_StartLBA),0);
		PGCCellADT[i].m_EndLBA = GetInt((uchar*)(&PGCCellADT[i].m_EndLBA),0);
	}

	// Now, go through our list and modify the appropriate entries.
	int j;
	for(j=0;j<NumEntries;j++)
	{
		// Find corresponding entry in existing table.
		for(i=0;i<OrigNumEntries;i++)
		{
			if(PGCCellADT[i].m_VOBID == VOBCellADT[j].m_VOBID && PGCCellADT[i].m_CELLID == VOBCellADT[j].m_CELLID)
				break;
		}
		if(i==OrigNumEntries)
		{
			// That's very strange, but it sometimes happens. The cell isn't declared in the table!
			// Let's see if we can add this entry without enlarging the table by 1 LBA...
			if((EndByte + sizeof(VTSM_C_ADTItem)) / SECTOR_SIZE == EndByte / SECTOR_SIZE)
			{
				ERROR3("Problem found in VTSM_C_ADT! Found a cell (%d/%d) that was not originally declared! Fixing the table...\n",
					VOBCellADT[j].m_VOBID,VOBCellADT[j].m_CELLID);
				ERRORF3("Problme dans VTSM_C_ADT! Une cellule a t trouve (%d/%d) qui n'est pas declare dans l'original! Table rpare...\n",
					VOBCellADT[j].m_VOBID,VOBCellADT[j].m_CELLID);
				PGCCellADT[i].m_StartLBA = VOBCellADT[j].m_StartLBA;
				PGCCellADT[i].m_EndLBA = VOBCellADT[j].m_EndLBA;
				PGCCellADT[i].m_VOBID = VOBCellADT[j].m_VOBID;
				PGCCellADT[i].m_CELLID = VOBCellADT[j].m_CELLID;
				OrigNumEntries++;
				EndByte += sizeof(VTSM_C_ADTItem);
				PutInt(Buffer,m_PVTSM_C_ADT*SECTOR_SIZE+4,EndByte);
				PutShort(Buffer,m_PVTSM_C_ADT*SECTOR_SIZE,(ushort)OrigNumEntries);
			}
			else
			{
				ERROR3("Problem in VTSM_C_ADT! Found a cell (%d/%d) that was not originally declared! Can't fix the table, you'll need an IFOEdit mock strip.\n",
					VOBCellADT[j].m_VOBID,VOBCellADT[j].m_CELLID);
				ERRORF3("Problme dans VTSM_C_ADT! Une cellule a t trouve (%d/%d) qui n'est pas declare dans l'original! Impossible de rparer la table, un Mock Strip IFOEdit est ncessaire.\n",
					VOBCellADT[j].m_VOBID,VOBCellADT[j].m_CELLID);
			}
		}
		else
		{
			// Found it. Correct sectors.
			PGCCellADT[i].m_StartLBA = VOBCellADT[j].m_StartLBA;
			PGCCellADT[i].m_EndLBA = VOBCellADT[j].m_EndLBA;
		}
	}

	// Convert back to right endian.
	for(i=0;i<OrigNumEntries;i++)
	{ 
		PGCCellADT[i].m_VOBID = GetShort((uchar*)(&PGCCellADT[i].m_VOBID),0);
		PGCCellADT[i].m_StartLBA = GetInt((uchar*)(&PGCCellADT[i].m_StartLBA),0);
		PGCCellADT[i].m_EndLBA = GetInt((uchar*)(&PGCCellADT[i].m_EndLBA),0);
	}
	return 0;
}

int CMenuShrinkEngine::CorrectVTSM_VOBU_ADMAP(uchar* Buffer, int NumEntries, uint* StartLBAs, 
											  int &OrigEndLBA, int &DeltaLBA)
{
	IFO_main*	MainIFO;
	uint			i;

	MainIFO = (IFO_main*)Buffer;

	
	// Find pointer to VTSM_VOBU_ADMAP
	uint m_PVTSM_VOBU_ADMAP = GetInt((uchar*)&MainIFO->m_PVTSM_VOBU_ADMAP,0);
	if(m_PVTSM_VOBU_ADMAP == 0)
		return 0;
	uint OrigNumEntries = (GetInt(Buffer,m_PVTSM_VOBU_ADMAP*SECTOR_SIZE)+1)/4-1;
	if(OrigNumEntries < (uint)NumEntries)
	{
		// This should never happen, but sometimes, the table is erroneous.
		// Let's see if we have enough room to grow the table... I've allocated MAX_IFO_GROWTH_LBA additional
		// sectors to grow the table. That should be enough in all cases!
		int GrowthLBA = (4*(NumEntries+1))/SECTOR_SIZE - (4*(OrigNumEntries+1))/SECTOR_SIZE;
		if(GrowthLBA > MAX_IFO_GROWTH_LBA)
		{
			// Not enough! Give up!
			ERROR3("The original VTSM_VOBU_ADMAP is incomplete. Found more VOBUs (%d) in the VOB file than declared \
in the original table (%d)\n",NumEntries,OrigNumEntries);
			ERRORF3("La table originale VTSM_VOBU_ADMAP est incomplete. Le nombre final de VOBUs (%d) est plus grand que celui dclar\
dans la table originale (%d)\n",NumEntries,OrigNumEntries);
			ERROR1("Can't fix the table, exiting\n");
			ERRORF1("Impossible de rparer la table...\n");
			// Can't fix the table. exiting...
			return 0;
		}
		else
		{
			if(GrowthLBA > 0)
			{
				// We'll grow the IFO. We need to move the stuff that's past us by GrowthLBA!
				OrigEndLBA = m_PVTSM_VOBU_ADMAP + (4*(OrigNumEntries+1))/SECTOR_SIZE + 1;
				memmove( Buffer+(OrigEndLBA+GrowthLBA)*SECTOR_SIZE, Buffer+OrigEndLBA*SECTOR_SIZE, 
					m_IFOFileSize - OrigEndLBA*SECTOR_SIZE);
			}
			ERROR3("The original VTSM_VOBU_ADMAP is incomplete. Found more VOBUs (%d) in the VOB file than declared \
in the original table (%d)\n",NumEntries,OrigNumEntries);
			ERRORF3("La table originale VTSM_VOBU_ADMAP est incomplete. Le nombre final de VOBUs (%d) est plus grand que celui dclar\
dans la table originale (%d)\n",NumEntries,OrigNumEntries);
			ERROR1("Fixing the table\n");
			ERRORF1("Table rpare...\n");
		}
	}
	uchar*	Array = Buffer+m_PVTSM_VOBU_ADMAP*SECTOR_SIZE;
	// Write new entries...
	for(i=0;i<(uint)NumEntries;i++)
	{
		PutInt(Array,4+i*4,StartLBAs[i]);
	}
	OrigEndLBA = m_PVTSM_VOBU_ADMAP + (4*(OrigNumEntries+1))/SECTOR_SIZE + 1;
	// Fill the rest with 0, just to be sure.
	for(;i<OrigNumEntries;i++)
	{
		PutInt(Array,4+i*4,0);
	}
	// And write new length.
	PutInt(Array,0,4*(NumEntries+1)-1);
	// Figure out if the table has shrunk to a point that it's one or more LBA smaller.
	DeltaLBA = (4*(OrigNumEntries+1))/SECTOR_SIZE - (4*(NumEntries+1))/SECTOR_SIZE;
	return 0;
}

// This function corrects pointers within an IFO file, according to modifications brought to the VOB.
// It does not correct pointers in video_ts.ifo that pertain to how various titleset are laid out on
// the disk. That's done in FinalGetVTSSectors();
int CMenuShrinkEngine::CorrectLocalVTSSectors(uchar* Buffer, int OrigEndLBA, int DeltaLBA, bool VMG)
{
	// First of, change the pointers to table that are beyond the end of the VTSM_VOBU_ADMAP table.
	IFO_main*	MainIFO;

	MainIFO = (IFO_main*)Buffer;

	if(DeltaLBA != 0)
	{
		// The IFO has shrunk by 1 LBA or more, because we reduced the size of VTSM_VOBU_ADMAP.
		// Move everything that's beyond it by DeltaLBA.
		if(DeltaLBA > 0)
		{
			memmove( Buffer+(OrigEndLBA-DeltaLBA)*SECTOR_SIZE, Buffer+OrigEndLBA*SECTOR_SIZE, 
				m_IFOFileSize - OrigEndLBA*SECTOR_SIZE);
		}

		uint PVTSM_VOBU_ADMAP = GetInt((uchar*)&MainIFO->m_PVTSM_VOBU_ADMAP ,0);
		if(VMG)
		{
			// Only for VIDEO_TS.IFO
			uint PTT_SRPT = GetInt((uchar*)&MainIFO->m_PTT_SRPT,0);
			if(PTT_SRPT > PVTSM_VOBU_ADMAP)
				PTT_SRPT -= DeltaLBA;
			PutInt((uchar*)&MainIFO->m_PTT_SRPT ,0,PTT_SRPT );
		}

		// If any of these pointers pointed past PVTSM_VOBU_ADMAP they should be decremented by however
		// many sectors the IFO file shrunk by.
		uint PVTS_PTT_SRPT = GetInt((uchar*)&MainIFO->m_PVTS_PTT_SRPT ,0);
		uint PVTS_PGCI = GetInt((uchar*)&MainIFO->m_PVTS_PGCI ,0);
		uint PVTSM_PGCI_UT = GetInt((uchar*)&MainIFO->m_PVTSM_PGCI_UT ,0);
		uint PVTS_TMAPTI = GetInt((uchar*)&MainIFO->m_PVTS_TMAPTI ,0);
		uint PVTSM_C_ADT = GetInt((uchar*)&MainIFO->m_PVTSM_C_ADT ,0);

		if(PVTS_PTT_SRPT > PVTSM_VOBU_ADMAP)
			PVTS_PTT_SRPT -= DeltaLBA;
		if(PVTS_PGCI > PVTSM_VOBU_ADMAP)
			PVTS_PGCI -= DeltaLBA;
		if(PVTSM_PGCI_UT > PVTSM_VOBU_ADMAP)
			PVTSM_PGCI_UT -= DeltaLBA;
		if(PVTS_TMAPTI > PVTSM_VOBU_ADMAP)
			PVTS_TMAPTI -= DeltaLBA;
		if(PVTSM_C_ADT > PVTSM_VOBU_ADMAP)
			PVTSM_C_ADT -= DeltaLBA;

		// Write these guys back.
		PutInt((uchar*)&MainIFO->m_PVTS_PTT_SRPT ,0,PVTS_PTT_SRPT );
		PutInt((uchar*)&MainIFO->m_PVTS_PGCI ,0,PVTS_PGCI );
		PutInt((uchar*)&MainIFO->m_PVTSM_PGCI_UT ,0,PVTSM_PGCI_UT );
		PutInt((uchar*)&MainIFO->m_PVTS_TMAPTI ,0,PVTS_TMAPTI );
		PutInt((uchar*)&MainIFO->m_PVTSM_C_ADT ,0,PVTSM_C_ADT );

		// Only for VTS_xx_0.IFO
		if(!VMG)
		{
			uint PVTS_C_ADT = GetInt((uchar*)&MainIFO->m_PVTS_C_ADT ,0);
			uint PVTS_VOBU_ADMAP = GetInt((uchar*)&MainIFO->m_PVTS_VOBU_ADMAP ,0);
			if(PVTS_C_ADT > PVTSM_VOBU_ADMAP)
				PVTS_C_ADT -= DeltaLBA;
			if(PVTS_VOBU_ADMAP > PVTSM_VOBU_ADMAP)
				PVTS_VOBU_ADMAP -= DeltaLBA;
			PutInt((uchar*)&MainIFO->m_PVTS_C_ADT ,0,PVTS_C_ADT );
			PutInt((uchar*)&MainIFO->m_PVTS_VOBU_ADMAP ,0,PVTS_VOBU_ADMAP );
		}
	}
	m_IFOFileSize -= DeltaLBA*SECTOR_SIZE;

	// Now set 1C, last sector of IFO file.
	PutInt(Buffer,0x1c,(m_IFOFileSize/SECTOR_SIZE)-1);
	// Set C0, first sector of VTS_xx_0.VOB.
	PutInt(Buffer,0xc0,m_IFOFileSize/SECTOR_SIZE);
	if(!VMG)
	{
		// Set C4, first sector of VTS_xx_1.VOB.
		// Find size of VTS_xx_0.VOB
		char Temp[CHAR_SIZE];
		sprintf(Temp,"%s0.VOB",m_BaseVTSName);
		int MenuSize = GetFinalFileSize(Temp);
		// Then find cumulative size of title vobs.
		int i;
		int TitleSize = 0;
		for(i=1;i<10;i++)
		{
			sprintf(Temp,"%s%d.VOB",m_BaseVTSName,i);
			TitleSize += GetFinalFileSize(Temp);
		}
		PutInt(Buffer,0xc4,(m_IFOFileSize/SECTOR_SIZE)+MenuSize);
		PutInt(Buffer,0x0c,2*(m_IFOFileSize/SECTOR_SIZE)+MenuSize+TitleSize-1);
	}
	else
	{
		// I am missing TT_SRPT, and for that I need all the VOBs! Done in FinalGetVTSSectors().
		char Temp[CHAR_SIZE];
		sprintf(Temp,"%s.VOB",m_BaseVTSName);
		int MenuSize = GetFinalFileSize(Temp);
		PutInt(Buffer,0x0c,2*(m_IFOFileSize/SECTOR_SIZE)+MenuSize-1);
	}
	return 0;	
}

// This function updates the TT_SRPT table. For this to work, video_ts.ifo must be loaded in Buffer.
int CMenuShrinkEngine::FinalGetVTSSectors()
{
	// Let's gather all the info we need. We need the sizes of the titlesets.
	int i,j;
	char Temp[CHAR_SIZE];

	// Load VIDEO_TS.IFO. First try the working version, in case VIDEO_TS.VOB was processed.
	sprintf(Temp,"%s/VIDEO_TS.IFO",m_WorkingDirFullPath);
	FILE* fd = fopen(Temp,"rb");
	if(fd == NULL)
	{
		// Apparently, video_ts was not processed, try opening the backup file.
		sprintf(Temp,"%s/VIDEO_TS.IFO",m_BackupDirFullPath);
		fd = fopen(Temp,"rb");
		if(fd == 0)
		{
			ERROR1("Could not open VIDEO_TS.IFO for update, exiting\n");
			ERRORF1("Impossible de modifier VIDEO_TS.IFO \n");
			return -1;
		}
	}
	fseek(fd,0,SEEK_END);
	m_IFOFileSize = ftell(fd);
	uchar* IFOFile = new uchar[m_IFOFileSize];
	fseek(fd,0,SEEK_SET);
	fread(IFOFile,1,m_IFOFileSize,fd);
	fclose(fd);

	m_VTSInfo[0].m_IFOFinalSize = m_IFOFileSize/SECTOR_SIZE;
	sprintf(Temp,"%s/VIDEO_TS.VOB",m_WorkingDirFullPath);
	m_VTSInfo[0].m_VTSMenuFinalSize = GetFinalFileSize(Temp);
	m_VTSInfo[0].m_VTSTitleSize = 0;
	uint VMGSize = 2*m_VTSInfo[0].m_IFOFinalSize+m_VTSInfo[0].m_VTSMenuFinalSize;

	// Then the vobs.
	uint StartOfVTS[100];
	StartOfVTS[0] = 0;
	for(i=1;i<100;i++)
	{
		sprintf(Temp,"%s/VTS_%02d_0.IFO",m_WorkingDirFullPath,i);
		m_VTSInfo[i].m_IFOFinalSize = GetFinalFileSize(Temp);
		sprintf(Temp,"%s/VTS_%02d_0.VOB",m_WorkingDirFullPath,i);
		m_VTSInfo[i].m_VTSMenuFinalSize = GetFinalFileSize(Temp);
		int Size = 0;
		for(j=1;j<10;j++)
		{
			sprintf(Temp,"%s/VTS_%02d_%d.VOB",m_WorkingDirFullPath,i,j);
			Size += GetFinalFileSize(Temp);
		}
		m_VTSInfo[i].m_VTSTitleSize = Size;
		StartOfVTS[i] = StartOfVTS[i-1]+2*m_VTSInfo[i-1].m_IFOFinalSize+m_VTSInfo[i-1].m_VTSMenuFinalSize+
			m_VTSInfo[i-1].m_VTSTitleSize;
	}

	// Find the TT_SRPT table.
	uint PTT_SRPT = GetInt(IFOFile,0xc4);

	// Get number of titles.
	uint NumTitles = GetShort(IFOFile,SECTOR_SIZE*PTT_SRPT);

	for(i=0;i<(int)NumTitles;i++)
	{
		// Get VTS number:
		uchar VTSNumber = IFOFile[SECTOR_SIZE*PTT_SRPT+8+i*12+6];
		PutInt(IFOFile,SECTOR_SIZE*PTT_SRPT+8+i*12+8,StartOfVTS[VTSNumber]);
	}

	sprintf(Temp,"%s/VIDEO_TS.IFO",m_WorkingDirFullPath);
	fd = fopen(Temp,"wb");
	if(fd == NULL)
	{
		ERROR2("Could not open %s for update, exiting\n",Temp);
		ERRORF2("Impossible de modifier %s\n",Temp);
		return -1;
	}
	fwrite(IFOFile,1,m_IFOFileSize,fd);
	fclose(fd);
	// Don't forget the BUP file!
	sprintf(Temp,"%s/VIDEO_TS.BUP",m_WorkingDirFullPath);
	fd = fopen(Temp,"wb");
	if(fd == NULL)
	{
		ERROR2("Could not open %s for update, exiting\n",Temp);
		ERROR2("Impossible de modifier %s\n",Temp);
		return -1;
	}
	fwrite(IFOFile,1,m_IFOFileSize,fd);
	fclose(fd);
	delete[] IFOFile;
	return 0;
}

int CMenuShrinkEngine::ScanDVDFolder(char* Path, bool ResetSkips)
{
	// Simple, split the path, find the folder. Try to find all the usual vob files. If you can't,
	// try to add VIDEO_TS to the path and try again!
	char dir[CHAR_SIZE];
	char fname[CHAR_SIZE];
	char ext[16];
	char Temp[CHAR_SIZE];

	
	// Maybe the path was a folder itself, try that: Ex: RETURN_OF_JEDI/VIDEO_TS
	if(ScanFolder(Path,ResetSkips) != -1)
	{
		_chdir(Path);
		sprintf(m_BackupDirFullPath,"%s/%s",Path,m_BackupDirName);
		strcpy(m_WorkingDirFullPath,Path);
		return 0;
	}
	// No, didn't work. Split path.
	SplitPath( Path, dir, fname, ext);
	// Try scanning the folder part of the path: Ex: RETURN_OF_JEDI/VIDEO_TS/video_ts.ifo
	if(ScanFolder(dir,ResetSkips) != -1)
	{
		_chdir(dir);
		sprintf(m_BackupDirFullPath,"%s%s",dir,m_BackupDirName);
		dir[strlen(dir)-1]='\0';
		strcpy(m_WorkingDirFullPath,dir);
		return 0;
	}
	// didn't find anything. Try to see if this is the parent of VIDEO_TS
	sprintf(Temp,"%s%s/VIDEO_TS",dir,fname);
	if(ScanFolder(Temp,ResetSkips) != -1)
	{
		_chdir(Temp);
		sprintf(m_BackupDirFullPath,"%s/%s",Temp,m_BackupDirName);
		strcpy(m_WorkingDirFullPath,Temp);
		return 0;
	}
	// Well we really didn't find anything. Too bad.
	strcpy(m_BackupDirFullPath,"");
	strcpy(m_WorkingDirFullPath,"");
	return -1;
}

// A helper function that determines the size of a file, looking first in the backup folder.
int CMenuShrinkEngine::GetOriginalFileSize(char* FileName)
{
	int Size;
	char Temp[CHAR_SIZE];
	// Try to open the file in backup folder..
	sprintf(Temp,"%s/%s",m_BackupDirName,FileName); 
	FILE* fd = fopen(Temp,"rb");
	if(fd == 0) 
	{
		// If you can't find it, try in the original folder.
		fd = fopen(FileName,"rb");
	}
	if(fd == 0)
		return -1;
	// My low-tech way of finding the size.
	fseek(fd,0,SEEK_END);
	Size =  ftell(fd)/SECTOR_SIZE;
	fclose(fd);
	return Size;
}

// A helper function that determines the final size of a file, not looking in the backup directory!
int CMenuShrinkEngine::GetFinalFileSize(char* FileName)
{
	int Size;
	// If you can't find it, try in the original folder.
	FILE* fd = fopen(FileName,"rb");
	if(fd == 0)
		return 0;
	// My low-tech way of finding the size.
	fseek(fd,0,SEEK_END);
	Size =  ftell(fd)/SECTOR_SIZE;
	fclose(fd);
	return Size;
}


int CMenuShrinkEngine::ScanFolder(char* Path, bool ResetSkips)
{
	// First thing you look for is VIDEO_TS.IFO.
	char Temp[CHAR_SIZE];
	int Size, i, j;
	int Ret = -1;
	// Erase last scan results.
	m_LastVTS = -1;
//	memset(m_VTSInfo,0,100*sizeof(VTSInfo));
	m_NumSectorsToProcess = 0;
	m_SectorsProcessedSoFar = 0;
	m_OrigCumulMenuSize = 0;
	// Try to _chdir into Path.
	if(strcmp(Path,"") != 0 && _chdir(Path) != 0)
		return -1;
	// We're there.
	Size = GetOriginalFileSize("VIDEO_TS.IFO");
	if(Size != -1)
	{
		// Found it! Try to find the VOB files.
		Ret = 1;
		if(ResetSkips) 
			m_VTSInfo[0].m_SkipThisVTS = 0;
		Size = GetOriginalFileSize("VIDEO_TS.VOB");
		if(Size != -1)
		{
			m_LastVTS = 1;
			m_VTSInfo[0].m_VTSMenuOrigSize = Size;
			m_NumSectorsToProcess = Size;
			m_OrigCumulMenuSize += Size;
		}
		else
			m_VTSInfo[0].m_VTSMenuOrigSize = 0;
		for(i=1;i<100;i++)
		{
			if(ResetSkips) 
				m_VTSInfo[i].m_SkipThisVTS = 0;
			sprintf(Temp,"VTS_%02d_0.VOB",i);
			Size = GetOriginalFileSize(Temp);
			if(Size != -1)
			{
				m_VTSInfo[i].m_VTSMenuOrigSize = Size;
				m_OrigCumulMenuSize += Size;
				if(m_VTSInfo[i].m_SkipThisVTS == 0)
					m_NumSectorsToProcess += m_VTSInfo[i].m_VTSMenuOrigSize;
				m_LastVTS = i;
			}
			else
				m_VTSInfo[i].m_VTSMenuOrigSize = 0;
			// Now determine the cumulative size of the title VOB files. We will use it in the
			// final GetVTS Sectors.
			int CumulSize = 0;
			for(j=1;j<10;j++)
			{
				sprintf(Temp,"VTS_%02d_%d.VOB",i,j);
				Size = GetOriginalFileSize(Temp);
				if(Size > 0)
					CumulSize += Size;
			}
			m_VTSInfo[i].m_VTSTitleSize = CumulSize;
		}
	}
	if(strcmp(Path,"") != 0)
		_chdir("..");
	return Ret;
}


void CMenuShrinkEngine::GetOveralDVDSize()
{
	m_OverallDVDSize = 0;
	for(int i=0;i<100;i++)
	{
		if(m_VTSInfo[i].m_VTSMenuFinalSize > 0)
			m_OverallDVDSize += m_VTSInfo[i].m_VTSMenuFinalSize;
		else
			m_OverallDVDSize += m_VTSInfo[i].m_VTSMenuOrigSize;
		m_OverallDVDSize += m_VTSInfo[i].m_VTSTitleSize;
	}
	MESSAGE3("Overall DVD Size %dMB (%.2fGB)\n",m_OverallDVDSize/512,(float)m_OverallDVDSize/512/1024);
	MESSAGEF3("Taille totale du DVD: %dMB (%.2fGB)\n",m_OverallDVDSize/512,(float)m_OverallDVDSize/512/1024);
}
