#ifndef MENUSHRINKENGINE_H
#define MENUSHRINKENGINE_H

#include <stdio.h>
#include "includes.h"
#include "Cell.h"
// Including IFOUtils for the definitions of the tables. Unfortunately, I coded IFOUtils after
// I coded MenuShrinkEngine, so the latter does not use the nice CIFOLoader class. 
#include "IFOUtils.h"

#define VERSION "MenuShrink Version 2.41"

#if 0
typedef enum
{
	FrameSelect_First,
	FrameSelect_Last,
	FrameSelect_Largest,
} FrameSelect;
#else
typedef int FrameSelect;
#endif



typedef struct
{
	// Some extra info I need to keep around.
	int		m_KeepAudio;		// Indicates whether the audio was kept for that cell.
	int		m_KillIt;			// Indicates this cell should be skipped if possible.
	int		m_KeepAnimated;		// Indicates whether the cell was kept animated.
	int		m_WasReferenced;	// Indicates if this cell was referenced in the original DVD.
	int		m_CellDuration90;	// Duration of cell measured as PTS (90kHz ticks).
	int		m_NumAudioStreams;	// Number of audio streams in that cell.
	uint	m_AudioStreams[8];	// Audio streams IDs.
	int		m_AnyLinkNext;
} CELL_EXTRA_INFO;

typedef struct
{
	// All sizes in sectors.
	int		m_VTSMenuOrigSize;
	int		m_VTSMenuFinalSize;
	int		m_IFOFinalSize;
	int		m_VTSTitleSize;
	int		m_SkipThisVTS;
} VTSInfo;

class CMenuShrinkEngine
{
public:

	CMenuShrinkEngine();
	~CMenuShrinkEngine();

	// The two CCell instances that hold the input and output cell data.
	CCell*		m_Cell1;
	CCell*		m_Cell2;

	bool		m_FixSubPicStops;		// Enables experimental code that removes "stops" in subpic data.
	bool		m_UseInfiniteStills;	// If set, 255 is used for still menus with no audio.
	int			m_OneSecondStill;		// If 1, cells that have 1 VOBU will be given a still of 1s.
	int			m_SkipButtonlessCells;	// If 1, buttonless cells are skipped, if possible.
	int			m_MinButtonSize;		// Cells with buttons smaller than this and now forced action are killed.
	int			m_PreviewAllCells;		// If true, the FrameSelectFunction() function will be called on all
										// cells that have more than 1 I-frame

	// Various arrays used in the IFOs.
	CELL_PLAYBACK_INFO* m_CellPBInfo;
	CELL_POS_INFO*		m_CellPosInfo;
	VTSM_C_ADTItem*		m_ADTInfo;
	CELL_EXTRA_INFO*	m_CellXInfo;
	int			m_NumAllocated;			// How many of the above arrays have we allocated.
	int			m_NumCells;				// How many actual arrays are present.

	// Another array used in the IFOs.
	uint*		m_NavPackLBATable;		
	int			m_NumAllocNavPacks;		// How big is the above table.
	int			m_NumNavPacks;			// Now many actual nav pack do we have in the array.

	char		m_BackupDirName[256];		// Default backup directory name (MenuShrinkBackup). 
	char		m_BackupDirFullPath[256];	// Full path to backup directory.
	char		m_WorkingDirFullPath[256];	// Full path to working directory.
	char		m_BaseVTSName[256];			// Part of the VTS file name, ex: VTS_01_ 
	char		m_InFileName[256];			// The VTS file name, ex: VTS_01_0.VOB

	VTSInfo		m_VTSInfo[100];			// Array of info on VTS files.
	int			m_LastVTS;				// Index of last VTS present.

	int			m_IFOFileSize;			// Size of the current IFO file in bytes.
	int			m_AbortNow;				// Set to 1 to abort processing.
	int			m_Busy;					// Signals we're in the middle of processing.

	FrameSelect m_FrameSelect;			// Which frame to select during shrinking.
	int			m_KeepAudio;			// Whether or not we keep the audio streams.

	int			m_SectorsProcessedSoFar;	// How many sectors have we processed thus far.
	int			m_NumSectorsToProcess;		// How many sectors do we have to process.
	int			m_OrigCumulMenuSize;		// Size in sectors of original menus.
	int			m_FinalCumulMenuSize;		// Size in sectors of processed menus.
	char*		m_VTSCurrentlyProcessing;
	int			m_CurrentVTSIndex;
	int			m_OverallDVDSize;		// Size of DVD in sectors, after processing.

	// This scans the directory in Path and gathers info about which VTS are present etc. Valid paths are:
	// "DUMBO" "DUMBO/VIDEO_TS" or "DUMBO/VIDEO_TS/VTS_01_0.VOB". If ResetSkips is true, the VTS skip flags
	// are reset to 0.
	int			ScanDVDFolder(char* Path, bool ResetSkips = false);
	void		GetOveralDVDSize();
	// This processes all the found VTS in the DVD, according to the info in m_VTSInfo.
	int			ProcessFullDVD();
	int			RestoreBackup();

	// This will shrink all the cell in a VTS menu, which can be passed as a filename, or as a number
	// where 0 stands for VIDEO_TS.VOB.
	int			ProcessVTS(char* VTSFileName, int KeepAudio, FrameSelect WhichFrame);
	int			ProcessVTS(int VTSNum, int KeepAudio, FrameSelect WhichFrame);

	// This compresses all the cells in the vob file and writes the result in OurFileName. If KeepAudio
	// is true, audio packs are preserved. If PickLastFrame is 1, the last I-Frame of each cell is chosen
	// as the still, instead of the first one. If PickLastFrame is -1, the largest one is chosen.
	int			ShrinkAllCells(char* InFileName, char* OutFileName, int KeepAudio, FrameSelect WhichFrame);

	// This compresses a given cells in the vob file and writes the result in OurFileName. If KeepAudio
	// is true, audio packs are preserved. If PickLastFrame is 1, the last I-Frame of each cell is chosen
	// as the still, instead of the first one. If PickLastFrame is -1, the largest one is chosen.
	int			ShrinkOneCell(char* InFileName, char* OutFileName, ushort TheVOBID, uchar TheCELLID, 
		int KeepAudio, FrameSelect WhichFrame);

	// This extracts a given cell in the vob file and writes the result in OurFileName.
	int			ExtractOneCell(char* InFileName, char* OutFileName, ushort TheVOBID, uchar TheCELLID);

	// This updates the IFO file according to the modifications we've made. Set VMG to 1 to indicate
	// the video_ts.ifo file!
	int			UpdateIFO(char* IFONameInput, char* IFONameOutput, bool VMG);

protected:

	// Helper functions.
	int			FinalGetVTSSectors();
	int			CorrectLocalVTSSectors(uchar* Buffer, int OrigEndLBA, int DeltaLBA, bool VMG);
	int			CorrectVTSM_VOBU_ADMAP(uchar* Buffer, int NumEntries, uint* StartLBAs, 
		int &OrigEndLBA, int &DeltaLBA);
	int			CorrectVTSM_C_ADT(uchar* Buffer, int NumEntries, VTSM_C_ADTItem* Items);
	int			CorrectPGCInfos(uchar* Buffer, uint NumEntries, CELL_PLAYBACK_INFO* ItemsPB, 
					CELL_POS_INFO* ItemsPos, bool VMG, CELL_EXTRA_INFO*	CellXInfo);
	int			GatherCellInfo(CCell* TheCell);

	int			ScanFolder(char* Path, bool ResetSkips = false);
	int			GetOriginalFileSize(char* FileName);
	int			GetFinalFileSize(char* FileName);

	void		ResetAll();
				
};


#endif